# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import os

import pytest
from flexmock import flexmock
from marshmallow import ValidationError
from ogr import GithubService, PagureService

from packit.config import (
    CommonPackageConfig,
    Config,
    JobConfig,
    JobConfigTriggerType,
    JobConfigView,
    JobType,
    PackageConfig,
)
from packit.config.aliases import DEFAULT_VERSION
from packit.schema import JobConfigSchema, JobMetadataSchema


def get_job_config_dict_simple(**update):
    result = {
        "job": "copr_build",
        "trigger": "release",
    }
    result.update(update)
    return result


def get_job_config_simple(**kwargs):
    """pass kwargs to JobConfig constructor"""
    package_name = kwargs.get("downstream_package_name", "package")
    return JobConfig(
        type=JobType.copr_build,
        trigger=JobConfigTriggerType.release,
        packages={package_name: CommonPackageConfig(**kwargs)},
    )


@pytest.fixture()
def job_config_simple():
    return get_job_config_simple()


def get_job_config_dict_full():
    return {
        "job": "propose_downstream",
        "trigger": "pull_request",
        "dist_git_branches": ["master"],
    }


def get_job_config_full(**kwargs):
    """pass kwargs to JobConfig constructor"""
    package_name = kwargs.get("downstream_package_name", "package")
    return JobConfig(
        type=JobType.propose_downstream,
        trigger=JobConfigTriggerType.pull_request,
        packages={
            package_name: CommonPackageConfig(dist_git_branches=["master"], **kwargs),
        },
    )


@pytest.fixture()
def job_config_full():
    return get_job_config_full()


def get_job_config_dict_build_for_branch():
    return {
        "job": "copr_build",
        "trigger": "commit",
        "branch": "build-branch",
        "scratch": True,
    }


def get_job_config_build_for_branch(**kwargs):
    """pass kwargs to JobConfig constructor"""
    package_name = kwargs.get("downstream_package_name", "package")
    return JobConfig(
        type=JobType.copr_build,
        trigger=JobConfigTriggerType.commit,
        packages={
            package_name: CommonPackageConfig(
                branch="build-branch",
                scratch=True,
                **kwargs,
            ),
        },
    )


def get_default_job_config(**kwargs):
    """pass kwargs to JobConfig constructor"""
    package_name = kwargs.get("downstream_package_name", "package")
    return [
        JobConfig(
            type=JobType.copr_build,
            trigger=JobConfigTriggerType.pull_request,
            packages={
                package_name: CommonPackageConfig(_targets=[DEFAULT_VERSION], **kwargs),
            },
        ),
        JobConfig(
            type=JobType.tests,
            trigger=JobConfigTriggerType.pull_request,
            packages={
                package_name: CommonPackageConfig(_targets=[DEFAULT_VERSION], **kwargs),
            },
        ),
        JobConfig(
            type=JobType.propose_downstream,
            trigger=JobConfigTriggerType.release,
            packages={
                package_name: CommonPackageConfig(
                    dist_git_branches=["fedora-all"],
                    **kwargs,
                ),
            },
        ),
    ]


def test_job_config_equal(job_config_simple):
    assert job_config_simple == job_config_simple


def test_job_config_not_equal(job_config_simple, job_config_full):
    assert job_config_simple != job_config_full


def test_job_config_blah():
    with pytest.raises(ValidationError) as ex:
        JobConfig.get_from_dict({"job": "asdqwe", "trigger": "salt"})
    assert "'trigger': ['Invalid enum member salt']" in str(ex.value)
    assert "'job': ['Invalid enum member asdqwe']" in str(ex.value)


def expanded_job_config_dict(**kwargs):
    config = {
        "job": "copr_build",
        "trigger": "pull_request",
        "packages": {"package": {"specfile_path": "package.spec"}},
    }
    config["packages"]["package"].update(**kwargs)
    return config


def expanded_job_config_object(**kwargs):
    package_config = {"specfile_path": "package.spec"}
    package_config.update(**kwargs)
    return JobConfig(
        type=JobType.copr_build,
        trigger=JobConfigTriggerType.pull_request,
        packages={"package": CommonPackageConfig(**package_config)},
    )


@pytest.mark.parametrize(
    "raw,is_valid",
    [
        ({}, False),
        ({"trigger": "release"}, False),
        ({"release_to": ["f28"]}, False),
        (expanded_job_config_dict(), True),
        (expanded_job_config_dict(dist_git_branches=["main"]), True),
    ],
)
def test_job_config_validate(raw, is_valid):
    if is_valid:
        JobConfig.get_from_dict(raw)
    else:
        with pytest.raises(ValidationError):
            JobConfig.get_from_dict(raw)


@pytest.mark.parametrize(
    "raw,expected_config",
    [
        (expanded_job_config_dict(), expanded_job_config_object()),
        (
            expanded_job_config_dict(dist_git_branches=["main"]),
            expanded_job_config_object(dist_git_branches=["main"]),
        ),
        (
            expanded_job_config_dict(),
            expanded_job_config_object(preserve_project=False),
        ),
        (
            expanded_job_config_dict(preserve_project=False),
            expanded_job_config_object(preserve_project=False),
        ),
        (
            expanded_job_config_dict(preserve_project=True),
            expanded_job_config_object(preserve_project=True),
        ),
    ],
)
def test_job_config_parse(raw, expected_config):
    job_config = JobConfig.get_from_dict(raw_dict=raw)
    assert job_config == expected_config


def test_deserialize_job_config_view():
    job_config = JobConfig(
        type=JobType.build,
        trigger=JobConfigTriggerType.commit,
        packages={
            "a": CommonPackageConfig(specfile_path="a.spec"),
            "b": CommonPackageConfig(specfile_path="b.spec"),
        },
    )

    job_config_view = JobConfigView(job_config, "a")
    assert job_config_view.identifier == "a"
    dump = JobConfigSchema().dump(job_config_view)
    job_config_view_deserialization = JobConfig.get_from_dict(dump)
    assert isinstance(job_config_view_deserialization, JobConfigView)
    assert job_config_view.identifier == "a"


@pytest.mark.parametrize(
    "raw,expected,allowed_pr_authors,allowed_committers",
    [
        pytest.param(
            {
                "job": "koji_build",
                "trigger": "commit",
                "packages": {"package": {"specfile_path": "package.spec"}},
            },
            JobConfig(
                type=JobType.koji_build,
                trigger=JobConfigTriggerType.commit,
                packages={"package": CommonPackageConfig(specfile_path="package.spec")},
            ),
            ["packit"],
            [],
        ),
        pytest.param(
            {
                "job": "koji_build",
                "trigger": "commit",
                "packages": {
                    "package": {
                        "specfile_path": "package.spec",
                        "allowed_committers": ["me"],
                    },
                },
            },
            JobConfig(
                type=JobType.koji_build,
                trigger=JobConfigTriggerType.commit,
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="package.spec",
                        allowed_committers=["me"],
                    ),
                },
            ),
            ["packit"],
            ["me"],
        ),
        pytest.param(
            {
                "job": "koji_build",
                "trigger": "commit",
                "packages": {
                    "package": {
                        "specfile_path": "package.spec",
                        "allowed_committers": ["me"],
                        "allowed_pr_authors": [],
                    },
                },
            },
            JobConfig(
                type=JobType.koji_build,
                trigger=JobConfigTriggerType.commit,
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="package.spec",
                        allowed_committers=["me"],
                        allowed_pr_authors=[],
                    ),
                },
            ),
            [],
            ["me"],
        ),
    ],
)
def test_koji_build_allowlist(raw, expected, allowed_pr_authors, allowed_committers):
    job_config = JobConfig.get_from_dict(raw_dict=raw)
    assert job_config == expected
    assert job_config.allowed_pr_authors == allowed_pr_authors
    assert job_config.allowed_committers == allowed_committers


@pytest.mark.parametrize(
    "raw,expected,allowed_builders",
    [
        pytest.param(
            {
                "job": "bodhi_update",
                "trigger": "commit",
                "packages": {
                    "package": {
                        "specfile_path": "package.spec",
                    },
                },
            },
            JobConfig(
                type=JobType.bodhi_update,
                trigger=JobConfigTriggerType.commit,
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="package.spec",
                        allowed_builders=["packit"],
                    ),
                },
            ),
            ["packit"],
        ),
        pytest.param(
            {
                "job": "bodhi_update",
                "trigger": "commit",
                "packages": {
                    "package": {
                        "specfile_path": "package.spec",
                        "allowed_builders": ["me"],
                    },
                },
            },
            JobConfig(
                type=JobType.bodhi_update,
                trigger=JobConfigTriggerType.commit,
                packages={
                    "package": CommonPackageConfig(
                        specfile_path="package.spec",
                        allowed_builders=["me"],
                    ),
                },
            ),
            ["me"],
        ),
    ],
)
def test_bodhi_updates_allowed(raw, expected, allowed_builders):
    job_config = JobConfig.get_from_dict(raw_dict=raw)
    assert job_config == expected
    assert job_config.allowed_builders == allowed_builders


@pytest.mark.parametrize(
    "raw,expected_packages_keys,identifiers",
    [
        pytest.param(
            {
                "upstream_project_url": "https://github.com/namespace/package",
                "packages": {
                    "package_a": {
                        "downstream_package_name": "package_a",
                        "paths": ["."],
                        "specfile_path": "package_a.spec",
                        "files_to_sync": ["package_a.spec", ".packit.yaml"],
                        "upstream_package_name": "package",
                    },
                    "package_b": {
                        "downstream_package_name": "package_b",
                        "identifier": "identifier_for_package_b",
                        "paths": ["."],
                        "specfile_path": "package_b.spec",
                        "files_to_sync": ["package_b.spec", ".packit.yaml"],
                        "upstream_package_name": "package",
                    },
                    "package_c": {
                        "downstream_package_name": "package_c",
                        "paths": ["."],
                        "specfile_path": "package_c.spec",
                        "files_to_sync": ["package_c.spec", ".packit.yaml"],
                        "upstream_package_name": "package_c",
                    },
                },
                "jobs": [
                    {
                        "job": "propose_downstream",
                        "trigger": "release",
                        "packages": {
                            "package_a": {"specfile_path": "a/package.spec"},
                            "package_b": {"specfile_path": "b/package.spec"},
                        },
                    },
                    {
                        "job": "propose_downstream",
                        "trigger": "release",
                        "packages": {
                            "package_c": {"specfile_path": "c/package.spec"},
                        },
                    },
                ],
            },
            [
                ["package_a"],
                ["package_b"],
                ["package_c"],
            ],
            ["package_a", "identifier_for_package_b", "package_c"],
        ),
        pytest.param(
            {
                "upstream_project_url": "https://github.com/namespace/package",
                "packages": {
                    "package_a": {
                        "downstream_package_name": "package_a",
                        "paths": ["."],
                        "specfile_path": "package_a.spec",
                        "files_to_sync": ["package_a.spec", ".packit.yaml"],
                        "upstream_package_name": "package",
                    },
                },
                "jobs": [
                    {
                        "job": "propose_downstream",
                        "trigger": "release",
                        "packages": {
                            "package_a": {"specfile_path": "a/package.spec"},
                        },
                    },
                ],
            },
            [
                ["package_a"],
            ],
            [None],
        ),
    ],
)
def test_job_config_views(raw, expected_packages_keys, identifiers):
    pkg_config = PackageConfig.get_from_dict(raw_dict=raw)
    job_views = pkg_config.get_job_views()
    for job_config_view, keys, identifier in zip(
        job_views,
        expected_packages_keys,
        identifiers,
    ):
        assert job_config_view.identifier == identifier
        assert set(job_config_view.packages.keys()) == set(keys)
        assert pkg_config.get_package_config_for(job_config_view)


def test_get_user_config(tmp_path):
    user_config_file_path = tmp_path / ".packit.yaml"
    user_config_file_path.write_text(
        "---\n"
        "debug: true\n"
        "fas_user: rambo\n"
        "keytab_path: './rambo.keytab'\n"
        "kerberos_realm: STG.FEDORAPROJECT.ORG\n"
        "github_token: GITHUB_TOKEN\n"
        "pagure_user_token: PAGURE_TOKEN\n",
    )
    flexmock(os).should_receive("getenv").with_args("XDG_CONFIG_HOME").and_return(
        str(tmp_path),
    )
    config = Config.get_user_config()
    assert config.debug and isinstance(config.debug, bool)
    assert config.fas_user == "rambo"
    assert config.keytab_path == "./rambo.keytab"
    assert config.kerberos_realm == "STG.FEDORAPROJECT.ORG"
    assert config.pkg_tool == "fedpkg"

    assert GithubService(token="GITHUB_TOKEN") in config.services
    assert PagureService(token="PAGURE_TOKEN") in config.services


def test_get_user_config_new_authentication(tmp_path):
    user_config_file_path = tmp_path / ".packit.yaml"
    user_config_file_path.write_text(
        "---\n"
        "debug: true\n"
        "fas_user: rambo\n"
        "keytab_path: './rambo.keytab'\n"
        "authentication:\n"
        "    github.com:\n"
        "        token: GITHUB_TOKEN\n"
        "    pagure:\n"
        "        token: PAGURE_TOKEN\n"
        '        instance_url: "https://my.pagure.org"\n',
    )
    flexmock(os).should_receive("getenv").with_args("XDG_CONFIG_HOME").and_return(
        str(tmp_path),
    )
    config = Config.get_user_config()
    assert config.debug and isinstance(config.debug, bool)
    assert config.fas_user == "rambo"
    assert config.keytab_path == "./rambo.keytab"
    assert config.kerberos_realm == "FEDORAPROJECT.ORG"

    assert GithubService(token="GITHUB_TOKEN") in config.services
    assert (
        PagureService(token="PAGURE_TOKEN", instance_url="https://my.pagure.org")
        in config.services
    )


def test_user_config_fork_token(tmp_path, recwarn):
    user_config_file_path = tmp_path / ".packit.yaml"
    user_config_file_path.write_text(
        "---\n" "pagure_fork_token: yes-is-true-in-yaml-are-you-kidding-me?\n",
    )
    flexmock(os).should_receive("getenv").with_args("XDG_CONFIG_HOME").and_return(
        str(tmp_path),
    )
    Config.get_user_config()
    w = recwarn.pop(UserWarning)
    assert "pagure_fork_token" in str(w.message)


@pytest.mark.parametrize(
    "config",
    [
        expanded_job_config_object(),
        expanded_job_config_object(dist_git_branches=["main"]),
    ],
)
def test_serialize_and_deserialize_job_config(config):
    schema = JobConfigSchema()
    serialized = schema.dump(config)
    new_config = schema.load(serialized)
    assert new_config == config


@pytest.mark.parametrize(
    "config_in,config_out,validation_error",
    [
        (
            {
                "job": "build",
                "trigger": "release",
                "packages": {
                    "package": {
                        "specfile_path": "packages.spec",
                        "enable_net": False,
                        "branch": "main",
                    },
                },
            },
            {
                "job": "build",
                "trigger": "release",
                "packages": {
                    "package": {
                        "specfile_path": "packages.spec",
                        "enable_net": False,
                        "branch": "main",
                    },
                },
            },
            False,
        ),
        (
            {
                "job": "build",
                "trigger": "release",
                "packages": {
                    "package": {
                        "specfile_path": "packages.spec",
                        "enable_net": False,
                        "branch": "main",
                    },
                },
            },
            {
                "job": "build",
                "trigger": "release",
                "packages": {
                    "package": {
                        "specfile_path": "packages.spec",
                        "enable_net": False,
                        "branch": "main",
                    },
                },
            },
            False,
        ),
        (
            {
                "job": "build",
                "trigger": "release",
                "packages": {
                    "package": {
                        "specfile_path": "packages.spec",
                        "enable_net": False,
                        "branch": "main",
                    },
                },
            },
            {
                "job": "build",
                "trigger": "release",
                "packages": {
                    "package": {
                        "specfile_path": "packages.spec",
                        "enable_net": False,
                        "branch": "main",
                    },
                },
            },
            False,
        ),
        (
            {
                "job": "tests",
                "trigger": "pull_request",
                "labels": ["regression", "long"],
                "packages": {
                    "package": {
                        "specfile_path": "packages.spec",
                        "enable_net": False,
                        "branch": "main",
                    },
                },
            },
            {
                "job": "tests",
                "trigger": "pull_request",
                "labels": ["regression", "long"],
                "packages": {
                    "package": {
                        "specfile_path": "packages.spec",
                        "enable_net": False,
                        "branch": "main",
                    },
                },
            },
            False,
        ),
        (
            {
                "job": "tests",
                "trigger": "ignore",
                "labels": ["regression", "long"],
                "packages": {
                    "package": {
                        "specfile_path": "packages.spec",
                        "enable_net": False,
                        "branch": "main",
                    },
                },
            },
            {
                "job": "tests",
                "trigger": "ignore",
                "labels": ["regression", "long"],
                "packages": {
                    "package": {
                        "specfile_path": "packages.spec",
                        "enable_net": False,
                        "branch": "main",
                    },
                },
            },
            False,
        ),
    ],
)
def test_deserialize_and_serialize_job_config(config_in, config_out, validation_error):
    schema = JobConfigSchema()
    if validation_error:
        with pytest.raises(ValidationError, match=validation_error):
            schema.dump(schema.load(config_in))
    else:
        serialized_from_in = schema.dump(schema.load(config_in))
        serialized_from_out = schema.dump(schema.load(config_out))
        assert serialized_from_in == serialized_from_out


@pytest.mark.parametrize(
    "config,is_valid",
    [
        ({}, True),
        ({"targets": []}, True),
        ({"targets": {}}, True),
        ({"targets": ["this", "is", "list"]}, True),
        ({"targets": {"a": {}, "b": {"distros": ["rhel-7"]}}}, True),
        ({"targets": {"a": {}, "b": {"additional_modules": ""}}}, True),
        (
            {
                "targets": {
                    "a": {},
                    "b": {
                        "additional_modules": "asd:1.2",
                        "additional_repos": ["http://foo.bar/"],
                    },
                },
            },
            True,
        ),
        (
            {"targets": {"b": {"additional_modules": None, "additional_repos": False}}},
            False,
        ),
        ({"targets": {"a": {}, "b": {"unknown": ["rhel-7"]}}}, False),
        ({"targets": {"a": {}, "b": {"distros": "not a list"}}}, False),
        ({"targets": {"this", "is", "set"}}, False),
    ],
)
def test_job_metadata_targets(config, is_valid):
    if is_valid:
        JobMetadataSchema().load(config)
    else:
        with pytest.raises(ValidationError):
            JobMetadataSchema().load(config)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
import os

import click

from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_packit_api, iterate_packages
from packit.config import get_context_settings, pass_config
from packit.constants import (
    PACKAGE_LONG_OPTION,
    PACKAGE_OPTION_HELP,
    PACKAGE_SHORT_OPTION,
)
from packit.utils.changelog_helper import ChangelogHelper

logger = logging.getLogger("packit")


@click.command("in-mock", context_settings=get_context_settings())
@click.option(
    "--upstream-ref",
    default=None,
    help="Git ref of the last upstream commit in the current branch "
    "from which packit should generate patches "
    "(this option implies the repository is source-git).",
)
@click.option(
    "--release-suffix",
    default=None,
    type=click.STRING,
    help="Specifies release suffix. Allows to override default generated:"
    "{current_time}.{sanitized_current_branch}{git_desc_suffix}",
)
@click.option(
    "--default-release-suffix",
    default=False,
    is_flag=True,
    help=(
        "Allows to use default, packit-generated, release suffix when some "
        "release_suffix is specified in the configuration."
    ),
)
@click.option(
    "-r",
    "--root",
    default="default",
    help=(
        "Uses specified chroot configuration. If ends with '.cfg', then it is "
        "treated as full path to the mock configuration."
    ),
)
@click.option(
    PACKAGE_SHORT_OPTION,
    PACKAGE_LONG_OPTION,
    multiple=True,
    help=PACKAGE_OPTION_HELP.format(action="build"),
)
@click.argument(
    "path_or_url",
    type=LocalProjectParameter(),
    default=os.path.curdir,
)
@pass_config
@cover_packit_exception
@iterate_packages
def mock(
    config,
    upstream_ref,
    release_suffix,
    default_release_suffix,
    root,
    package_config,
    path_or_url,
):
    """
    Build RPMs in mock using content of the upstream repository.

    PATH_OR_URL argument is a local path or a URL to the upstream git repository,
    it defaults to the current working directory.
    """
    api = get_packit_api(
        config=config,
        package_config=package_config,
        local_project=path_or_url,
    )
    release_suffix = ChangelogHelper.resolve_release_suffix(
        api.package_config,
        release_suffix,
        default_release_suffix,
    )

    if not config.srpm_path:
        config.srpm_path = api.create_srpm(
            upstream_ref=upstream_ref,
            srpm_dir=api.up.local_project.working_dir,
            release_suffix=release_suffix,
        )

    rpm_paths = api.run_mock_build(root=root, srpm_path=config.srpm_path)
    logger.info("RPMs:")
    for path in rpm_paths:
        logger.info(f" * {path}")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import configparser
import logging
import os
from pathlib import Path
from typing import Union

import pyrpkg

from packit.exceptions import PackitLookasideCacheException

logger = logging.getLogger(__name__)


class LookasideCache:
    def __init__(self, pkg_tool: str) -> None:
        """
        Constructs a wrapper around a lookaside cache.

        Args:
            pkg_tool: Packaging tool associated with a lookaside cache instance
            to be used to get the URLs.

        Raises:
            PackitLookasideCacheException, if construction of the wrapper fails,
            e.g can't read or parse the config, or missing settings are required.
        """
        try:
            parser = configparser.ConfigParser()
            parser.read(f"/etc/rpkg/{pkg_tool}.conf")
            self._config = dict(parser.items(pkg_tool, raw=True))
        except configparser.Error as e:
            raise PackitLookasideCacheException(
                "Failed to parse the rpkg config",
            ) from e

        try:
            self.cache = pyrpkg.lookaside.CGILookasideCache(
                self._config["lookasidehash"],
                self._config["lookaside"],
                self._config["lookaside_cgi"],
            )
        except KeyError as e:
            raise PackitLookasideCacheException(
                "Failed to create a CGI for lookaside cache",
            ) from e

    def _get_package(self, package: str) -> str:
        if self._config.get("lookaside_namespaced", False):
            return f"rpms/{package}"
        return package

    def get_sources(
        self,
        basepath: Union[Path, str],
        package: str,
    ) -> list[dict[str, str]]:
        """
        Gets URLs to sources stored in lookaside cache.

        Args:
            basepath: Path to a dist-git repo containing the "sources" file.
            package: Package name.

        Returns:
            List of dicts with path (filename) and URL.

        Raises:
            PackitLookasideCacheException, if parsing the "sources" file fails.
        """
        try:
            sources = pyrpkg.sources.SourcesFile(Path(basepath) / "sources", "bsd")
        except (pyrpkg.errors.MalformedLineError, ValueError) as e:
            raise PackitLookasideCacheException("Failed to parse sources") from e
        result = []
        for entry in sources.entries:
            url = self.cache.get_download_url(
                self._get_package(package),
                entry.file,
                entry.hash,
                entry.hashtype,
            )
            result.append({"path": entry.file, "url": url})
        return result

    def is_archive_uploaded(self, package: str, archive_path: Union[Path, str]) -> bool:
        """
        We are using a name to check the presence in the lookaside cache.
        (This is the same approach fedpkg itself uses.)
        """
        archive_name = os.path.basename(archive_path)
        archive_hash = self.cache.hash_file(archive_path)

        return self.cache.remote_file_exists_head(
            name=self._get_package(package),
            filename=archive_name,
            hash=archive_hash,
            # `hashtype` is a required argument, yet it doesn't have a default…
            hashtype=None,
        )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import logging

import pytest
from flexmock import flexmock
from requests import HTTPError, Session

from packit.exceptions import ImageBuilderError, PackitException
from packit.vm_image_build import ImageBuilder


def test_create_image():
    """Test creating an image"""
    # auth is tested in tests_recording
    flexmock(ImageBuilder).should_receive("refresh_auth").and_return(None)
    ib = ImageBuilder("foo-token")
    payload = {
        "image_name": "mona-lisa",
        "distribution": "rhel-90",
        "image_requests": [
            {
                "architecture": "x86_64",
                "image_type": "aws",
                "upload_request": {
                    "options": {"share_with_accounts": ["123456789012"]},
                    "type": "aws",
                },
            },
        ],
        "customizations": {
            "payload_repositories": [
                {
                    "rhsm": False,
                    "baseurl": "https://copr.fedoraproject.org/coprs/foo/bar/",
                    "check_gpg": False,
                },
            ],
            "packages": ["mona-lisa"],
        },
    }
    request_response = {"id": "foo-baz-bar"}
    flexmock(ib).should_receive("image_builder_request").with_args(
        "POST",
        "compose",
        payload=payload,
    ).and_return(flexmock(json=lambda: request_response))
    response = ib.create_image(
        "rhel-90",
        "mona-lisa",
        {
            "architecture": "x86_64",
            "image_type": "aws",
            "upload_request": {
                "options": {"share_with_accounts": ["123456789012"]},
                "type": "aws",
            },
        },
        {"packages": ["mona-lisa"]},
        "https://copr.fedoraproject.org/coprs/foo/bar/",
    )
    assert response == "foo-baz-bar"


def test_refresh_fails():
    """Refereshing console.rh.c access token should have sensible experience"""
    # We generate refresh token in the webui for packit and then use it to generate access token
    # Refresh token can expire: when that happens, all requests end with 401
    # Before this commit, we did not expect that to happen
    # The traceback we got took us some time to figure out what's happening exactly
    # This test ensures that Packit raises sensible exception when access token cannot be obtained
    flexmock(ImageBuilder).should_receive("_get_access_token").and_return(None)
    with pytest.raises(PackitException) as exc:
        ImageBuilder("this-token-definitely-not-works-unless...")
        assert (
            "Unable to obtain access token. You may need to regenerate the refresh token."
            in str(exc)
        )


def test_bad_request(caplog):
    def _raise_for_status(*args, **_):
        raise HTTPError()

    response_json = {
        "errors": [
            {
                "detail": (
                    "request body has an error: doesn't match schema... "
                    'value "fedora-rawhide" is not one of the allowed values'
                ),
            },
        ],
    }

    r = flexmock(
        ok=False,
        status_code=400,
        text=str(response_json),
        raise_for_status=_raise_for_status,
    )
    flexmock(ImageBuilder).should_receive("refresh_auth").and_return(None)
    flexmock(Session).should_receive("request").and_return(r)
    with caplog.at_level(logging.INFO, logger="packit"):
        ib = ImageBuilder("foobar")
        with pytest.raises(ImageBuilderError):
            ib.create_image("fedora-rawhide", "foo-bar", {}, {}, "")
        assert "request body has an error" in caplog.text
        assert 'value "fedora-rawhide" is not one of the allowed values' in caplog.text



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import logging
from datetime import datetime, timedelta

from ogr.abstract import Release

from packit.config import Config
from packit.config.common_package_config import MultiplePackages
from packit.copr_helper import CoprHelper
from packit.distgit import DistGit
from packit.exceptions import PackitException
from packit.upstream import Upstream
from packit.utils.bodhi import get_bodhi_client
from packit.utils.koji_helper import KojiHelper

logger = logging.getLogger(__name__)


class Status:
    """
    This class provides methods to obtain status of the package
    """

    def __init__(
        self,
        config: Config,
        package_config: MultiplePackages,
        upstream: Upstream,
        distgit: DistGit,
    ):
        self.config = config
        self.package_config = package_config

        self.up = upstream
        self.dg = distgit

    def get_downstream_prs(self, number_of_prs: int = 5) -> list[tuple[int, str, str]]:
        """
        Get specific number of latest downstream PRs
        :param number_of_prs: int
        :return: List of downstream PRs
        """
        table: list[tuple[int, str, str]] = []
        pr_list = self.dg.local_project.git_project.get_pr_list()
        logger.debug("Downstream PRs fetched.")
        if len(pr_list) > 0:
            # take last `number_of_prs` PRs
            pr_list = (
                pr_list[:number_of_prs] if len(pr_list) > number_of_prs else pr_list
            )
            table = [(pr.id, pr.title, pr.url) for pr in pr_list]
        return table

    def get_dg_versions(self) -> dict:
        """
        Get versions from all branches in Dist-git
        :return: Dict {"branch": "version"}
        """
        dg_versions = {}
        branches = self.dg.local_project.git_project.get_branches()
        logger.debug("Dist-git branches fetched.")
        for branch in branches:
            try:
                self.dg.create_branch(
                    branch,
                    base=f"remotes/origin/{branch}",
                    setup_tracking=False,
                )
                self.dg.switch_branch(branch)
                self.dg.specfile.reload()
            except Exception as ex:
                logger.debug(f"Branch {branch!r} is not present: {ex!r}.")
                continue
            try:
                dg_versions[branch] = self.dg.specfile.expanded_version
            except AttributeError:
                logger.debug(f"Can't figure out the version of branch: {branch}.")
        self.dg.switch_branch()

        return dg_versions

    def get_up_releases(self, number_of_releases: int = 5) -> list:
        """
        Get specific number of latest upstream releases
        :param number_of_releases: int
        :return: List
        """
        if self.up.local_project is None or self.up.local_project.git_project is None:
            logger.info("We couldn't track any upstream releases.")
            return []

        latest_releases: list[Release] = []
        try:
            latest_releases = self.up.local_project.git_project.get_releases()
            logger.debug("Upstream releases fetched.")
        except PackitException as e:
            logger.debug(f"Failed to fetch upstream releases: {e}")

        return latest_releases[:number_of_releases]

    def get_koji_builds(self) -> dict[str, str]:
        """Get latest koji builds as a dict of branch: latest build in that branch."""
        # This method returns only latest builds,
        # so we don't need to get whole build history from Koji,
        # get just recent year to speed things up.
        since = datetime.now() - timedelta(days=365)
        builds = KojiHelper().get_nvrs(
            self.dg.package_config.downstream_package_name,
            since,
        )
        logger.debug(f"Recent Koji builds fetched: {builds}")
        # Select latest build for each branch.
        # ['python-ogr-0.5.0-1.fc29', 'python-ogr-0.6.0-1.fc29']
        # -> {'fc29': 'python-ogr-0.6.0-1.fc29'}
        return {b.rsplit(".", 1)[1]: b for b in reversed(builds)}

    def get_updates(self, number_of_updates: int = 3) -> list:
        """
        Get specific number of latest updates in bodhi
        :param number_of_updates: int
        :return: None
        """
        bodhi_client = get_bodhi_client()
        updates = []
        page = pages = 1
        while page <= pages:
            results = bodhi_client.query(
                packages=self.dg.package_config.downstream_package_name,
                page=page,
            )
            updates.extend(results["updates"])
            page += 1
            pages = results["pages"]
        logger.debug("Bodhi updates fetched.")

        stable_branches: set[str] = set()
        all_updates = [
            [
                update["title"],
                update["karma"],
                update["status"],
                update["release"]["branch"],
            ]
            for update in updates
        ]
        updates = []
        for [update, karma, status, branch] in all_updates:
            # Don't return more than one stable update per branch
            if branch not in stable_branches or status != "stable":
                updates.append([update, karma, status])
                if status == "stable":
                    stable_branches.add(branch)
            if len(updates) == number_of_updates:
                break
        return updates

    def get_copr_builds(self, number_of_builds: int = 5) -> list:
        return (
            CoprHelper(upstream_local_project=self.up.local_project).get_copr_builds(
                number_of_builds=number_of_builds,
            )
            if self.up.local_project
            else []
        )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from pathlib import Path

import pytest

from packit.api import PackitAPI
from packit.exceptions import PackitConfigException
from packit.utils.commands import cwd


@pytest.mark.parametrize(
    "existing_files,existing_directories,raw_package_config,valid,expected_output",
    [
        (
            ["a.md", "b.md"],
            ["./a_dir"],
            """\
config_file_path: .packit.yaml
dist_git_base_url: https://packit.dev/
downstream_package_name: packit
upstream_ref: last_commit
upstream_package_name: packit_upstream
allowed_gpg_keys:
- gpg
dist_git_namespace: awesome
files_to_sync:
- .packit.yaml
- packit.spec
- src: a.md
  dest: aaa.md
- a_dir
            """,
            True,
            ".packit.yaml is valid and ready to be used",
        ),
        (
            ["a.md", "b.md"],
            ["./a_dir"],
            """\
config_file_path: .packit.yaml
dist_git_base_url: https://packit.dev/
downstream_package_name: packit
upstream_ref: last_commit
upstream_package_name: packit_upstream
allowed_gpg_keys:
- gpg
dist_git_namespace: awesome
files_to_sync:
- .packit.yaml
- packit.spec
- a.md
- b.md
- c.txt
- a_dir
            """,
            False,
            "The following path is configured to be synced but is not present "
            "in the repository: c.txt",
        ),
        (
            ["a.md"],
            [],
            """\
config_file_path: .packit.yaml
dist_git_base_url: https://packit.dev/
downstream_package_name: packit
upstream_ref: last_commit
upstream_package_name: packit_upstream
allowed_gpg_keys:
- gpg
dist_git_namespace: awesome
files_to_sync:
- .packit.yaml
- packit.spec
- a.md
- b.md
- c.txt
            """,
            False,
            "The following paths are configured to be synced but are not present "
            "in the repository: b.md, c.txt",
        ),
        (
            ["a.md", "./a_dir/file"],
            ["./a_dir"],
            """\
config_file_path: .packit.yaml
dist_git_base_url: https://packit.dev/
downstream_package_name: packit
upstream_ref: last_commit
upstream_package_name: packit_upstream
allowed_gpg_keys:
- gpg
dist_git_namespace: awesome
files_to_sync:
- .packit.yaml
- packit.spec
- a.md
- b.md
- c.txt
- ./a_dir/*
            """,
            False,
            "The following paths are configured to be synced but are not present "
            "in the repository: b.md, c.txt",
        ),
        (
            ["a.md", "b.md"],
            ["./a_dir"],
            """\
config_file_path: .packit.yaml
dist_git_base_url: https://packit.dev/
downstream_package_name: packit
upstream_ref: last_commit
upstream_package_name: packit_upstream
allowed_gpg_keys:
- gpg
dist_git_namespace: awesome
files_to_sync:
- .packit.yaml
- packit.spec
- ./a_dir/*
- a.md
- b.md
- c.txt
            """,
            False,
            "The following paths are configured to be synced but are not present "
            "in the repository: ./a_dir/*, c.txt",
        ),
        (
            ["a.md", "b.md"],
            ["./a_dir"],
            """\
config_file_path: .packit.yaml
dist_git_base_url: https://packit.dev/
downstream_package_name: packit
upstream_ref: last_commit
upstream_package_name: packit_upstream
allowed_gpg_keys:
- gpg
dist_git_namespace: awesome
files_to_sync:
- .packit.yaml
- packit.spec
- ./a_dir
- a.md
- b.md
            """,
            True,
            ".packit.yaml is valid and ready to be used",
        ),
    ],
    ids=[
        "none_missing",
        "one_missing",
        "two_missing",
        "dir_with_globs",
        "empty_dir_with_globs",
        "empty_dir",
    ],
)
def test_validate_paths(
    tmpdir,
    existing_files,
    existing_directories,
    raw_package_config,
    valid,
    expected_output,
):
    with cwd(tmpdir):
        Path(".packit.yaml").write_text(raw_package_config)
        Path("packit.spec").write_text("hello")
        for existing_directory in existing_directories:
            Path(existing_directory).mkdir()
        for existing_file in existing_files:
            Path(existing_file).write_text("")

        if valid:
            output = PackitAPI.validate_package_config(Path("."))
            assert expected_output in output
        else:
            with pytest.raises(PackitConfigException) as exc:
                PackitAPI.validate_package_config(Path("."))
                assert expected_output in exc



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Code that backs up Image Builder integration.

Guides and more info about Image Builder:
    https://access.redhat.com/documentation/en-us/red_hat_enterprise_linux/8/html/composing_a_customized_rhel_system_image/index
    https://www.redhat.com/en/blog/using-hosted-image-builder-its-api
    https://github.com/packit/research/tree/main/image-builder
"""
import logging
from typing import Optional, Union

import requests
from requests import HTTPError

from packit.exceptions import ImageBuilderError, PackitException

logger = logging.getLogger("packit")
REDHAT_SSO_AUTH_URL = (
    "https://sso.redhat.com/auth/realms/redhat-external/protocol/openid-connect/token"
)
REDHAT_API_GET_USER_URL = "https://api.access.redhat.com/account/v1/user"
API_ROOT = "https://console.redhat.com/api/"
IMAGE_BUILDER_API_URL = f"{API_ROOT}image-builder/v1"
LAUNCH_API_URL = f"{API_ROOT}provisioning/v1"


class ImageBuilder:
    """
    Client class to work with Image Builder API and Red Hat SSO.
    """

    def __init__(
        self,
        refresh_token: str,
    ):
        """
        Args:
            refresh_token: your personal token to access Red Hat SSO, obtain at:
                https://access.redhat.com/management/api
        """
        self.refresh_token = refresh_token
        self._access_token: Optional[str] = None

        self.session = requests.Session()
        self.refresh_auth()

    def refresh_auth(self):
        """
        Refresh the access token and the image_builder_session headers.
        """
        self._access_token = self._get_access_token()
        if not self._access_token:
            raise PackitException(
                "Unable to obtain access token. You may need to regenerate the refresh token.",
            )
        self.session.headers.update(
            {
                "Authorization": f"Bearer {self._access_token}",
                "Accept": "application/json",
            },
        )

    def request(self, method: str, url: str, payload: Optional[dict] = None):
        """
        Request to the Image Builder API.

        Args:
            method: HTTP method to use (GET, POST, PUT, DELETE)
            url: complete URL for the request
            payload: payload to send with the request

        Returns:
            requests.Response object.
        """
        # don't try to refresh twice
        token_is_fresh = False
        while True:
            response = self.session.request(method, url, json=payload)
            if response.status_code == 401 and not token_is_fresh:
                token_is_fresh = True
                self.refresh_auth()
                continue
            try:
                response.raise_for_status()
            except HTTPError as ex:
                logger.error(f"Image Builder Errors: {response.text}")
                raise ImageBuilderError(errors=response.text) from ex
            return response

    def image_builder_request(
        self,
        method: str,
        path: str,
        payload: Optional[dict] = None,
    ):
        """
        Request to the Image Builder API.

        Args:
            method: HTTP method to use (GET, POST, PUT, DELETE)
            path: path part in the URL to the API endpoint
            payload: payload to send with the request

        Returns:
            requests.Response object.
        """
        return self.request(
            method,
            url=f"{IMAGE_BUILDER_API_URL}/{path}",
            payload=payload,
        )

    def launch_request(self, method: str, path: str, payload: Optional[dict] = None):
        """
        Request to the Launch API.

        Args:
            method: HTTP method to use (GET, POST, PUT, DELETE)
            path: path part in the URL to the API endpoint
            payload: payload to send with the request

        Returns:
            requests.Response object.
        """
        return self.request(method, url=f"{LAUNCH_API_URL}/{path}", payload=payload)

    def _get_access_token(self) -> Optional[str]:
        """
        Obtain access token from Red Hat SSO using the provided refresh token.

            curl -X POST -d grant_type=refresh_token \
              -d client_id=rhsm-api -d refresh_token=$o \
              https://sso.redhat.com/auth/realms/redhat-external/protocol/openid-connect/token

        Returns:
            access token (used for interacting with the API)
        """
        sso = requests.Session()
        response = sso.post(
            REDHAT_SSO_AUTH_URL,
            data={
                "grant_type": "refresh_token",
                "client_id": "rhsm-api",
                "refresh_token": self.refresh_token,
            },
        )
        response_json = response.json()
        try:
            return response_json["access_token"]
        except KeyError:
            logger.info(
                f"Failed to get access token ({response.status_code}): {response_json}",
            )
            return None

    def check_token(self) -> dict:
        """
        Using the obtained access token, check that it is valid by obtaining the
        user info.

        Returns:
            Dict payload with the user info.
        """
        response = requests.get(
            REDHAT_API_GET_USER_URL,
            headers={"Authorization": f"Bearer {self._access_token}"},
        )
        return response.json()

    def create_image(
        self,
        image_distribution: str,
        image_name: str,
        image_request: dict,
        image_customizations: dict,
        repo_url: str,
    ):
        """
        Create an image using the Image Builder API.
        Images are called a compose in context of Image Builder.

        request and customization fields are passed to Image Builder, see
        their API for the definition, example:

          "customizations": {
            "filesystem": [
              {
                "min_size": 1024,
                "mountpoint": "/var"
              }
            ],
            "packages": [ "postgresql" ],
          }
          "image_requests": [
            {
              "architecture": "x86_64",
              "image_type": "aws",
              "upload_request": {
                "options": {
                  "share_with_accounts": [
                    "123456789012"
                  ]
                },
                "type": "aws"
              }
            }
          ]

        Args:
            image_distribution: distribution name (base operating system), e.g. "rhel-8"
            image_name: image name, e.g. "packit-test-john-foo"
            image_request: Image request definition of an image build
            image_customizations: Image customizations definition of an image build
            repo_url: yum repository URL, e.g.
                "https://download.copr.fedorainfracloud.org/" +
                "results/@cockpit/cockpit-preview/centos-stream-9-x86_64/"
        """
        image_customizations.setdefault("payload_repositories", [])
        # variable assignment is done for sake of mypy:
        #   https://stackoverflow.com/a/54788883/909579
        payload_repositories: list[dict[str, Union[str, bool]]] = image_customizations[
            "payload_repositories"
        ]
        logger.debug(
            f"image_customizations -> payload_repositories {payload_repositories}",
        )
        payload = {
            "image_name": image_name,
            "distribution": image_distribution,
            "image_requests": [image_request],
            "customizations": image_customizations,
        }
        payload_repositories.append(
            {
                "rhsm": False,
                "baseurl": repo_url,
                # needs to be the actual key, not link:
                # https://issues.redhat.com/browse/HMSIB-14
                # "gpgkey": "https://download.copr.../@cockpit/cockpit-preview/pubkey.gpg",
                "check_gpg": False,
            },
        )
        response = self.image_builder_request("POST", "compose", payload=payload)

        response_json = response.json()
        try:
            return response_json["id"]
        except KeyError as e:
            logger.error(
                f"Failed to create image ({response.status_code}): {response_json}",
            )
            raise PackitException(f"Failed to create image: {response_json}") from e

    def get_image_status(self, build_id: str):
        """
        Get image build status.

        Args:
            build_id: Build ID of the image.

        Returns:
            Status of the build (example: success, building, pending)
        """
        # Examples of /composes/{build_id}
        # {'image_status': {'status': 'building'}}
        # {'image_status': {
        #     'status': 'success', 'upload_status': {
        #         'options': {
        #             'ami': 'ami-08ee0689a0547e7ea',
        #             'region': 'us-east-1'
        #         },
        #         'status': 'success', 'type': 'aws'
        #     }
        # }
        response_json = self.image_builder_request("GET", f"composes/{build_id}").json()
        logger.debug(f"Image build metadata: {response_json}")
        return response_json["image_status"]["status"]

    def clone(self, build_id: str, region: str, share_with_accounts: list[str]):
        """
        Clone existing image into a new AWS region.

        Args:
            build_id: Build ID of the image.
            region: AWS region
            share_with_accounts: List of str of accounts to share the new image with

        Returns:
            Metadata of the clone build
        """
        response_json = self.image_builder_request(
            "POST",
            f"composes/{build_id}/clone",
            payload={
                "region": region,
                "share_with_accounts": share_with_accounts,
            },
        ).json()
        logger.debug(f"Image clone metadata: {response_json}")
        return response_json

    def get_clone(self, build_id: str):
        """
        Get image clone status.

        Args:
            build_id: Build ID of the clone image.

        Returns:
            Metadata of the clone build
        """
        response_json = self.image_builder_request("GET", f"/clones/{build_id}").json()
        logger.debug(f"Image build metadata: {response_json}")
        return response_json



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
packit started as source-git and we're making a source-git module after such a long time, weird
"""

import importlib.resources
import logging
import textwrap
from pathlib import Path
from typing import Optional

import git
import yaml
from specfile import Specfile

from packit.config import Config
from packit.constants import (
    DISTRO_DIR,
    FROM_DIST_GIT_TOKEN,
    REPO_NOT_PRISTINE_HINT,
    RPM_MACROS_FOR_PREP,
    SRC_GIT_CONFIG,
)
from packit.exceptions import PackitException
from packit.patches import PatchMetadata
from packit.pkgtool import PkgTool
from packit.utils import (
    commit_message_file,
    get_default_branch,
    get_file_author,
    run_command,
)
from packit.utils.lookaside import LookasideCache
from packit.utils.repo import is_git_repo, is_the_repo_pristine

logger = logging.getLogger(__name__)


# https://stackoverflow.com/questions/13518819/avoid-references-in-pyyaml
# mypy hated the suggestion from the SA ^, hence an override like this
class SafeDumperWithoutAliases(yaml.SafeDumper):
    def ignore_aliases(self, data):
        # no aliases/anchors in the dumped yaml text
        return True


class SourceGitGenerator:
    """
    Set up a source-git repo in an upstream repo taking downstream patches
    from the corresponding package in Fedora/CentOS/RHEL.

    Attributes:
        config: Packit configuration.
        source_git: Git repository to be initialized as a source-git repo.
        dist_git: Dist-git repository to be used for initialization.
        upstream_ref: Upstream ref which is going to be the starting point of the
            source-git history. This can be a branch, tag or commit sha. It is expected
            that the current HEAD and this ref point to the same commit.
        upstream_url: Git URL to be saved in the source-git configuration. If not specified,
            the fetch URL of the 'origin' remote is used, unless 'upstream_remote' is set.
        upstream_remote: Name of the remote from which the fetch URL is taken as the Git URL
            of the upstream project to be saved in the source-git configuration.
        pkg_tool: Packaging tool to be used to interact with the dist-git repo.
        pkg_name: Name of the package in dist-git.
    """

    def __init__(
        self,
        config: Config,
        source_git: git.Repo,
        dist_git: git.Repo,
        upstream_ref: str,
        upstream_url: Optional[str] = None,
        upstream_remote: Optional[str] = None,
        pkg_tool: Optional[str] = None,
        pkg_name: Optional[str] = None,
    ):
        self.config = config
        self.source_git = source_git
        self.dist_git = dist_git
        self.upstream_ref = upstream_ref
        try:
            self.upstream_url = upstream_url or next(
                self.source_git.remote(upstream_remote or "origin").urls,
            )
        except ValueError as exc:
            logger.debug(f"Failed when getting upstream URL: {exc}")
            raise PackitException(
                "Unable to tell the URL of the upstream repository because there is no remote "
                f"called '{upstream_remote or 'origin'}' in {self.source_git.working_dir}. "
                "Please specify the correct upstream remote using '--upstream-remote' or the "
                "upstream URL, using '--upstream-url'.",
            ) from exc
        self.pkg_tool = pkg_tool
        self.pkg_name = pkg_name or Path(self.dist_git.working_dir).name

        self._dist_git_specfile: Optional[Specfile] = None
        self.distro_dir = Path(self.source_git.working_dir, DISTRO_DIR)
        self._dist_git: Optional[git.Repo] = None
        self._patch_comments: dict = {}

    @property
    def dist_git_specfile(self) -> Specfile:
        if not self._dist_git_specfile:
            path = str(Path(self.dist_git.working_dir, f"{self.pkg_name}.spec"))
            self._dist_git_specfile = Specfile(
                path,
                sourcedir=self.dist_git.working_dir,
                autosave=True,
            )
        return self._dist_git_specfile

    def _run_prep(self):
        """
        run `rpmbuild -bp` in the dist-git repo to get a git-repo
        in the %prep phase so we can pick the commits in the source-git repo
        """
        files = importlib.resources.files("packit.data")
        with importlib.resources.as_file(
            files.joinpath("_packitpatch"),
        ) as _packitpatch:
            logger.info(f"expanding %prep section in {self.dist_git.working_dir}")

            rpmbuild_args = [
                "rpmbuild",
                "--nodeps",
                "--define",
                f"_packitpatch {_packitpatch}",
                "--define",
                f"_topdir {self.dist_git.working_dir!s}",
                "-bp",
                "--define",
                f"_specdir {self.dist_git.working_dir}",
                "--define",
                f"_sourcedir {self.dist_git.working_dir}",
            ]

            rpmbuild_args += RPM_MACROS_FOR_PREP
            if logger.level <= logging.DEBUG:  # -vv can be super-duper verbose
                rpmbuild_args.append("-v")
            rpmbuild_args.append(str(self.dist_git_specfile.path))

            run_command(
                rpmbuild_args,
                cwd=self.dist_git.working_dir,
                print_live=True,
            )

    def get_BUILD_dir(self):
        path = Path(self.dist_git.working_dir)
        build_dirs = [
            d
            for d in (path / "BUILD").iterdir()
            # ignore the specparts dir
            if d.is_dir() and not d.name.endswith("-SPECPARTS")
        ]
        if len(build_dirs) > 1:
            raise RuntimeError(
                f"More than one matching directory found in {path / 'BUILD'}",
            )
        if not build_dirs:
            raise RuntimeError(f"No subdirectory found in {path / 'BUILD'}")
        build_dir = build_dirs[0]
        # since rpm 4.20 the actual build dir is one level deeper
        if build_dir.name.endswith("-build"):
            build_dirs = [
                d
                for d in (path / "BUILD" / build_dir).iterdir()
                if d.is_dir() and d.name != "SPECPARTS"
            ]
            if len(build_dirs) > 1:
                raise RuntimeError(
                    f"More than one matching directory found in {path / 'BUILD' / build_dir}",
                )
            if not build_dirs:
                raise RuntimeError(
                    f"No subdirectory found in {path / 'BUILD' / build_dir}",
                )
            build_dir = build_dirs[0]
        return build_dir

    def _rebase_patches(self):
        """Rebase current branch against the from_branch."""
        to_branch = "dist-git-commits"  # temporary branch to store the dist-git history
        BUILD_dir = self.get_BUILD_dir()
        if not is_git_repo(BUILD_dir):
            raise RuntimeError(
                "Running %prep section wasn't successful. "
                "Make sure the package uses %autosetup.",
            )
        prep_repo = git.Repo(BUILD_dir)
        from_branch = get_default_branch(prep_repo)
        logger.info(f"Rebase patches from dist-git {from_branch}.")
        self.source_git.git.fetch(BUILD_dir, f"+{from_branch}:{to_branch}")

        # transform into {patch_name: patch_id}
        with self.dist_git_specfile.patches() as patches:
            patch_ids = {p.filename: p.number for p in patches}
            patch_comments = {p.filename: p.comments.raw for p in patches}

        # -2 - drop first commit which represents tarball unpacking
        # -1 - reverse order, HEAD is last in the sequence
        patch_commits = list(prep_repo.iter_commits(from_branch))[-2::-1]

        for commit in patch_commits:
            self.source_git.git.cherry_pick(
                commit.hexsha,
                keep_redundant_commits=True,
                allow_empty=True,
                strategy_option="theirs",
            )

            # Annotate commits in the source-git repo with patch_id. This info is not provided
            # during the rpm patching process so we need to do it here.
            metadata = PatchMetadata.from_git_trailers(commit)
            trailers = [("Patch-id", patch_ids[metadata.name])]
            patch_status = ""
            for line in patch_comments.get(metadata.name, []):
                patch_status += f"    # {line}\n"
            if patch_status:
                trailers.append(("Patch-status", f"|\n{patch_status}"))
            trailers.append((FROM_DIST_GIT_TOKEN, self.dist_git.head.commit.hexsha))

            author = None
            # If the commit subject matches the one used in _packitpatch
            # when applying patches with 'patch', get the original (first)
            # author of the patch file in dist-git.
            if commit.message.startswith(f"Apply patch {metadata.name}"):
                author = get_file_author(self.dist_git, metadata.name)
            logger.debug(f"author={author}")

            with commit_message_file(
                commit.message,
                trailers=trailers,
            ) as commit_message:
                self.source_git.git.commit(
                    file=commit_message,
                    author=author,
                    amend=True,
                    allow_empty=True,
                )

        self.source_git.git.branch("-D", to_branch)

    def _populate_distro_dir(self):
        """Copy files used in the distro to package and test the software to .distro.

        Raises:
            PackitException, if the dist-git repository is not pristine, that is,
            there are changed or untracked files in it.
        """
        if not is_the_repo_pristine(self.dist_git):
            raise PackitException(
                "Cannot initialize a source-git repo. "
                "The corresponding dist-git repository at "
                f"{self.dist_git.working_dir!r} is not pristine. "
                f"{REPO_NOT_PRISTINE_HINT}",
            )

        command = ["rsync", "--archive", "--delete"]
        for exclude in ["*.patch", "sources", ".git*"]:
            command += ["--filter", f"exclude {exclude}"]

        command += [
            str(self.dist_git.working_dir) + "/",
            str(self.distro_dir),
        ]

        self.distro_dir.mkdir(parents=True)
        run_command(command)

    def _reset_gitignore(self):
        reset_rules = textwrap.dedent(
            """\
            # Reset gitignore rules
            !*
            """,
        )
        Path(self.distro_dir, ".gitignore").write_text(reset_rules)

    def _configure_syncing(self):
        """Populate source-git.yaml"""
        with self.dist_git_specfile.patches() as patches:
            try:
                patch_id_digits = patches[0].number_digits
            except (IndexError, AttributeError):
                patch_id_digits = 1
        package_config = {
            "upstream_project_url": self.upstream_url,
            "upstream_ref": self.upstream_ref,
            "downstream_package_name": self.pkg_name,
            "specfile_path": f"{DISTRO_DIR}/{self.pkg_name}.spec",
            "patch_generation_ignore_paths": [DISTRO_DIR],
            "patch_generation_patch_id_digits": patch_id_digits,
            "sync_changelog": True,
            "files_to_sync": [
                {
                    # The trailing-slash is important, as we want to
                    # sync the content of the directory, not the directory
                    # as a whole.
                    "src": f"{DISTRO_DIR}/",
                    "dest": ".",
                    "delete": True,
                    "filters": [
                        "protect .git*",
                        "protect sources",
                        f"exclude {SRC_GIT_CONFIG}",
                        "exclude .gitignore",
                    ],
                },
            ],
        }
        if lookaside_sources := LookasideCache(
            self.pkg_tool or self.config.pkg_tool,
        ).get_sources(
            basepath=self.dist_git.working_dir,
            package=self.pkg_name,
        ):
            package_config["sources"] = lookaside_sources

        Path(self.distro_dir, SRC_GIT_CONFIG).write_text(
            yaml.dump(
                package_config,
                Dumper=SafeDumperWithoutAliases,
                default_flow_style=False,
                sort_keys=False,
            ),
        )

    def create_from_upstream(self):
        """Create a source-git repo, by transforming downstream patches
        in Git commits applied on top of the selected upstream ref.
        """
        upstream_ref_sha = self.source_git.git.rev_list("-1", self.upstream_ref)
        if upstream_ref_sha != self.source_git.head.commit.hexsha:
            raise PackitException(
                f"{self.upstream_ref!r} is not pointing to the current HEAD "
                f"in {self.source_git.working_dir!r}.",
            )
        with self.dist_git_specfile.prep() as prep:
            if "%autosetup" not in prep:
                raise PackitException(
                    "Initializing source-git repos for packages "
                    "not using %autosetup is not supported.",
                )

        self._populate_distro_dir()
        self._reset_gitignore()
        self._configure_syncing()

        spec = Specfile(
            f"{self.distro_dir}/{self.pkg_name}.spec",
            sourcedir=self.dist_git.working_dir,
            autosave=True,
        )
        with spec.patches() as patches:
            patches.clear()

        self.source_git.git.stage(DISTRO_DIR, force=True)
        message = f"""Initialize as a source-git repository

{FROM_DIST_GIT_TOKEN}: {self.dist_git.head.commit.hexsha}
"""
        self.source_git.git.commit(message=message)

        pkg_tool = PkgTool(
            fas_username=self.config.fas_user,
            directory=self.dist_git.working_dir,
            tool=self.pkg_tool or self.config.pkg_tool,
        )
        pkg_tool.sources()
        self._run_prep()
        self._rebase_patches()



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from packit.utils.commands import run_command


def remove_gpg_key_pair(gpg_binary: str, fingerprint: str):
    run_command(
        [gpg_binary, "--batch", "--yes", "--delete-secret-and-public-key", fingerprint],
    )



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import datetime
import io
import shutil
import subprocess
import tarfile
from collections.abc import Iterator
from pathlib import Path
from typing import Optional

import git
import pytest
from flexmock import flexmock
from gnupg import GPG
from ogr.abstract import PRStatus
from ogr.read_only import PullRequestReadOnly
from ogr.services.github import GithubProject, GithubService
from ogr.services.pagure import PagureProject, PagureService, PagureUser

from packit.api import PackitAPI
from packit.base_git import PackitRepositoryBase
from packit.cli.utils import get_packit_api
from packit.config import config, get_local_package_config
from packit.constants import FROM_SOURCE_GIT_TOKEN
from packit.distgit import DistGit
from packit.local_project import CALCULATE, LocalProject, LocalProjectBuilder
from packit.pkgtool import PkgTool
from packit.upstream import GitUpstream
from packit.utils.commands import cwd
from packit.utils.repo import create_new_repo
from tests.integration.utils import remove_gpg_key_pair
from tests.spellbook import (
    DATA_DIR,
    DISTGIT,
    NAME_VERSION,
    SOURCEGIT_SOURCEGIT,
    SOURCEGIT_UPSTREAM,
    TARBALL_NAME,
    UPSTREAM,
    get_test_config,
    git_add_and_commit,
    initiate_git_repo,
)

DOWNSTREAM_PROJECT_URL = "https://src.fedoraproject.org/not/set.git"
UPSTREAM_PROJECT_URL = "https://github.com/also-not/set.git"
SOURCE_GIT_RELEASE_TAG = "0.1.0"
HELLO_RELEASE = "1.0.1"


@pytest.fixture()
def mock_remote_functionality_upstream(upstream_and_remote, distgit_and_remote):
    u, _ = upstream_and_remote
    d, _ = distgit_and_remote
    return mock_remote_functionality(d, u)


@pytest.fixture()
def mock_remote_functionality_upstream_fast_forward_merge_branches(
    upstream_and_remote,
    distgit_and_remote,
):
    u, _ = upstream_and_remote
    d, _ = distgit_and_remote
    return mock_remote_functionality(d, u, fast_forward_merge_branch=True)


@pytest.fixture()
def mock_remote_functionality_downstream_autochangelog(
    upstream_and_remote,
    distgit_with_autochangelog_and_remote,
):
    u, _ = upstream_and_remote
    d, _ = distgit_with_autochangelog_and_remote
    return mock_remote_functionality(d, u)


@pytest.fixture()
def mock_remote_functionality_sourcegit(sourcegit_and_remote, distgit_and_remote):
    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    return mock_remote_functionality(upstream=sourcegit, distgit=distgit)


def mock_spec_download_remote_s(
    repo_path: Path,
    spec_dir_path: Optional[Path] = None,
    archive_ref: str = "HEAD",
    files_to_create=None,
):
    files_to_create = files_to_create or []
    spec_dir_path = spec_dir_path or repo_path

    def mock_download_remote_sources(*_):
        """mock download of the remote archive and place it into dist-git repo"""

        archive_cmd = [
            "git",
            "archive",
            "--output",
            str(spec_dir_path / TARBALL_NAME),
            "--prefix",
            f"{NAME_VERSION}/",
            archive_ref,
        ]
        subprocess.check_call(archive_cmd, cwd=repo_path)
        for f in files_to_create:
            subprocess.check_call(["touch", f], cwd=repo_path)

    flexmock(PackitRepositoryBase, download_remote_sources=mock_download_remote_sources)


def mock_remote_functionality(
    distgit: Path,
    upstream: Path,
    fast_forward_merge_branch=False,
):
    example_pr = PullRequestReadOnly(
        title="",
        id=42,
        status=PRStatus.open,
        url="",
        description="",
        author="",
        source_branch="",
        target_branch="",
        created=datetime.datetime(1969, 11, 11, 11, 11, 11, 11),
    )

    def mocked_create_pr(*args, **kwargs):
        return example_pr

    release = flexmock(body="Some description of the upstream release", url="some-url")
    flexmock(config).should_receive("get_project").and_return(
        flexmock(
            get_release=lambda name, tag_name: release,
        ),
    )
    flexmock(GithubService)
    github_service = GithubService()
    flexmock(
        GithubService,
        get_project=lambda repo, namespace: GithubProject(
            "also-not",
            github_service,
            "set",
            github_repo=flexmock(),
        ),
    )
    flexmock(
        PagureProject,
        get_git_urls=lambda: {"git": DOWNSTREAM_PROJECT_URL},
        fork_create=lambda: None,
        get_fork=lambda: PagureProject("", "", PagureService()),
        create_pr=mocked_create_pr,
        default_branch="main",
    )
    if fast_forward_merge_branch:
        flexmock(PagureProject).should_receive("create_pr").times(2).and_return(
            example_pr,
        )
    flexmock(
        GithubProject,
        get_git_urls=lambda: {"git": UPSTREAM_PROJECT_URL},
        fork_create=lambda: None,
        get_release=lambda: flexmock(url="url"),
    )
    flexmock(PagureUser, get_username=lambda: "packito")

    dglp = LocalProjectBuilder().build(
        working_dir=distgit,
        git_repo=CALCULATE,
        git_url="https://packit.dev/rpms/beer",
        git_service=PagureService(),
    )
    flexmock(
        DistGit,
        push_to_fork=lambda *args, **kwargs: None,
        # let's not hammer the production lookaside cache webserver
        is_archive_in_lookaside_cache=lambda archive_path: False,
        local_project=dglp,
    )
    flexmock(DistGit).should_receive("existing_pr").and_return(None)

    def mocked_new_sources(sources=None, offline=False):
        sources = sources or []
        if not all(Path(s).is_file() for s in sources):
            raise RuntimeError("archive does not exist")

    flexmock(PkgTool, new_sources=mocked_new_sources)
    flexmock(PackitAPI, init_kerberos_ticket=lambda: None)
    return upstream, distgit


@pytest.fixture()
def mock_patching():
    flexmock(GitUpstream).should_receive("create_patches").and_return(["patches"])
    flexmock(DistGit).should_receive("specfile_add_patches").with_args(["patches"])


@pytest.fixture()
def cwd_upstream(upstream_and_remote) -> Iterator[Path]:
    upstream, _ = upstream_and_remote
    with cwd(upstream):
        yield upstream


@pytest.fixture()
def sourcegit_and_remote(tmp_path):
    sourcegit_remote = tmp_path / "source_git_remote"
    sourcegit_remote.mkdir()
    create_new_repo(sourcegit_remote, ["--bare"])

    sourcegit_dir = tmp_path / "source_git"
    shutil.copytree(SOURCEGIT_UPSTREAM, sourcegit_dir)
    initiate_git_repo(sourcegit_dir, tag=SOURCE_GIT_RELEASE_TAG)
    subprocess.check_call(
        ["cp", "-R", SOURCEGIT_SOURCEGIT, tmp_path],
        cwd=sourcegit_remote,
    )
    git_add_and_commit(directory=sourcegit_dir, message="sourcegit content")

    return sourcegit_dir, sourcegit_remote


@pytest.fixture()
def source_git_repo(sourcegit_and_remote):
    source_git_dir, _ = sourcegit_and_remote
    return git.Repo(source_git_dir)


@pytest.fixture()
def dist_git_repo(distgit_and_remote):
    dist_git_dir, _ = distgit_and_remote
    return git.Repo(dist_git_dir)


@pytest.fixture()
def hello_source_git_repo(tmp_path):
    repo_dir = tmp_path / "src" / "hello"
    shutil.copytree(DATA_DIR / "src" / "hello", repo_dir)
    repo = git.Repo.init(repo_dir)
    repo.git.add(".")
    repo.git.commit(message="Initial commit")
    repo.git.tag(HELLO_RELEASE)
    repo.create_remote("origin", "https://example.com/hello.git")
    return repo


@pytest.fixture()
def hello_dist_git_repo(tmp_path):
    repo_dir = tmp_path / "rpms" / "hello"
    shutil.copytree(DATA_DIR / "rpms" / "hello", repo_dir)
    repo = git.Repo.init(repo_dir)
    repo.git.add(".")
    repo.git.commit(message="Initial commit", author="Engin Eer <eer@redhat.com>")
    return repo


@pytest.fixture()
def downstream_n_distgit(tmp_path):
    d_remote = tmp_path / "downstream_remote"
    d_remote.mkdir()
    create_new_repo(d_remote, ["--bare"])

    d = tmp_path / "dist_git"
    shutil.copytree(DISTGIT, d)
    initiate_git_repo(d, tag="0.0.0")

    u = tmp_path / "upstream_git"
    shutil.copytree(UPSTREAM, u)
    initiate_git_repo(u, push=False, upstream_remote=str(d_remote))

    return u, d


@pytest.fixture()
def upstream_instance(upstream_and_remote, tmp_path):
    with cwd(tmp_path):
        u, _ = upstream_and_remote
        c = get_test_config()

        pc = get_local_package_config(str(u))
        pc.upstream_project_url = str(u)
        lp = LocalProjectBuilder().build(working_dir=u, git_repo=CALCULATE)

        ups = GitUpstream(c, pc, lp)
        yield u, ups


@pytest.fixture()
def upstream_instance_with_two_commits(upstream_instance):
    u, ups = upstream_instance
    new_file = u / "new.file"
    new_file.write_text("Some content")
    git_add_and_commit(u, message="Add new file")
    return u, ups


@pytest.fixture()
def distgit_instance(
    upstream_and_remote,
    distgit_and_remote,
    mock_remote_functionality_upstream,
):
    u, _ = upstream_and_remote
    d, _ = distgit_and_remote
    c = get_test_config()
    pc = get_local_package_config(str(u))
    pc.upstream_project_url = str(u)
    dg = DistGit(c, pc, clone_path=str(d))
    return d, dg


@pytest.fixture()
def distgit_instance_with_autochangelog(
    upstream_and_remote,
    distgit_with_autochangelog_and_remote,
    mock_remote_functionality_downstream_autochangelog,
):
    u, _ = upstream_and_remote
    d, _ = distgit_with_autochangelog_and_remote
    c = get_test_config()
    pc = get_local_package_config(str(u))
    pc.upstream_project_url = str(u)
    dg = DistGit(c, pc, clone_path=str(d))
    return d, dg


@pytest.fixture()
def api_instance(upstream_and_remote, distgit_and_remote):
    u, _ = upstream_and_remote
    d, _ = distgit_and_remote

    c = get_test_config()
    api = get_packit_api(
        config=c,
        local_project=LocalProjectBuilder().build(
            working_dir=Path.cwd(),
            git_repo=CALCULATE,
        ),
    )
    return u, d, api


def mock_api_for_source_git(
    sourcegit: Path,
    distgit: Path,
    up_local_project: LocalProject,
):
    with cwd(sourcegit):
        c = get_test_config()
        pc = get_local_package_config(
            package_config_path=sourcegit / ".distro" / "source-git.yaml",
        )
        pc.upstream_project_url = str(sourcegit)
        return PackitAPI(c, pc, up_local_project, dist_git_clone_path=str(distgit))


@pytest.fixture()
def api_instance_source_git(sourcegit_and_remote, distgit_and_remote):
    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote
    builder = LocalProjectBuilder()
    up_lp = builder.build(working_dir=sourcegit, git_repo=CALCULATE)
    up_lp.git_project = (
        flexmock()
        .should_receive("get_release")
        .with_args(tag_name="0.1.0", name="0.1.0")
        .and_return(flexmock(url="some-url"))
        .mock()
    )
    return mock_api_for_source_git(sourcegit, distgit, up_lp)


def add_source_git_commit_trailer(api: PackitAPI):
    # The version in dg is different from up, sync it
    api.dg.specfile.version = api.up.specfile.expanded_version
    api.dg.commit(
        "Update spec",
        "",
        trailers=[
            (
                FROM_SOURCE_GIT_TOKEN,
                api.up.local_project.git_repo.head.commit.hexsha,
            ),
        ],
    )
    return api


@pytest.fixture()
def api_instance_update_source_git(api_instance_source_git):
    add_source_git_commit_trailer(api_instance_source_git)
    return api_instance_source_git


@pytest.fixture()
def api_instance_sync_push(sourcegit_and_remote, distgit_and_remote):
    sourcegit, _ = sourcegit_and_remote
    distgit, _ = distgit_and_remote

    repo = flexmock(
        default_branch="main",
        get_pulls=lambda state, sort, direction: [],
        fork="fork",
    )
    service = flexmock(user=flexmock(get_username=lambda: "packit"))
    up_lp = LocalProjectBuilder().build(
        working_dir=sourcegit,
        git_repo=CALCULATE,
        git_project=GithubProject(
            repo="beer",
            service=service,
            namespace="packit.dev",
            github_repo=repo,
        ),
    )
    api = mock_api_for_source_git(sourcegit, distgit, up_lp)

    add_source_git_commit_trailer(api)

    api.dg.specfile.version = "1.2.3"
    api.dg.commit("dist-git commit to be sync back", "")
    return api


@pytest.fixture()
def gnupg_instance() -> GPG:
    return GPG()


@pytest.fixture()
def private_gpg_key() -> str:
    """
    gnupg_instance.export_keys(key.fingerprint, secret=True, expect_passphrase=False)
    """

    return """-----BEGIN PGP PRIVATE KEY BLOCK-----

lQOYBFzRgy8BCADbrsCtPWqVWTNkl3U1LK3YWLnKiZIIMELv305s6Lfj8lTtFlAT
GXuIfalAqN18Rv6h+/aMXW872Gwk6Iv9hdkSF9e04YGqdrr7H/rw8976NogvhR73
aTi1BGh43AtFWifJy/MaSCVfy2gTeYqK17FHDnzqGunwQ4L0PQMvkscMOgmCQlPj
qF2WWiknku/aMeIoqAjUfIV3/dKVYPJH7g4QTo8U00CvfLSFMFhgzxW/nC+fPOh6
h7jXnXk2xvo+9rWPrPplUtjjuufjzxvz/azlB3PIR05tTRU27P2xI4rJBkZdzZ6t
67obFxImpRVJtV3kvyBPAja4k7x/mWQvC/dBABEBAAEAB/kB8DWKgcV4OmCB9XUn
CjUheMzw3MxhTp20lJ2SR+5hcECwE9eSh5HHt0YgSC0mHNE/2COJgwSJfGQd4kBj
9QOgjX3NfoTgnmoRb6uM5zXzMrp6YtwOVksWC8spL9XYn45E0UwckgDkarzJGTQv
++24QQg4n5KrWEkmQwiNaafgc3lyFf+xaCri2xlwMYdFqltROzrckHkcbjYFdASr
2dqoxGn79OdRhndg8+n1FA2UpQhI4fZyvQwkfO7x1Mjl3DxH97K6cvnW+KI0IlhE
2laZRowzqn8q8+zopWtFkhQmD/SV43eLfCbzyb0KKyAheH8zD6DS7Ij/KkcnLmTt
pMzxBADb0DkUb/mAArhzBDf4QBup+ryhuae9cPkhMlVKiegtojlmmsdTKgCxybFH
M33ITeNx84IvDTZCZkHFwyFWVXLTbQyIt7RsGOEYxbt7cI+SgmtvEdYtnf6Y1H7F
0WksC8ES4Z6BToQ4qeI5rd3QAk+qmQ4ZSA9iT7vDG8/9Wv0TKQQA/9kFB318C39+
m1W6m9/B482brEJqrGaAkFT4yOSjeo1C+n3b/iedCROwP44L1ifZT83uJ3ad/o/f
N5iDHiXjASVnIuuehLjuwrauZhkhiOcjRyLRtrDweF5NDQu70o69ON3j4fWebhmT
OFxfGaD9lkWuM2Lf0/0kbErc8X31nlkD/2tTepbT6y1ud2OrI9Fw4E8eKnSycJwW
JitKqllXkpugaEihBtpAf7zcarchnF2FIqnYuT5nVZ067lPKU5rRMuD3D+IXUxSy
aNgCSzVMX2t43DQcyZ35kqgLYOJwdc78tNtkNvbTkZBRDTpcdK69M6L+x+wMsHb1
PpLui3F4WAQ/Q2O0IUF1dG9nZW5lcmF0ZWQgS2V5IDxwYWNraXRAcGFja2l0PokB
TgQTAQgAOBYhBFvV9i+3z2ur7IY5eKtE1ItYjzIoBQJc0YMvAhsvBQsJCAcCBhUK
CQgLAgQWAgMBAh4BAheAAAoJEKtE1ItYjzIoaIQH/RQ47hZhyGz9vgD196KIUwTp
WLrJPVxNSd4mqx0lwE3B5T8xyboZHZoD5gNxFR/6CPs2Nh4fiyqjKzeU/t5W4Y3c
evyhgBJu9y4K1s7HHf+Sby0jlaeQyVs11Ngoul+CM2m6ZzlLyexEC8dUZ9fclDxb
TOQH3GkJ24vdkbZwN+KdL/AYtbRAvE2BwfK0EMg7ibRoh9Zfpc2hYLjBZ83yAKgY
FZ8bkeRu7lTdzpbTu/nEFKKDYusgbJuLBaW3GEjj726C/IHAp16QZI/SPKpt0cAK
YEWYFA0MxyZQhRqEDH2whr+QyWr5155N7kzHnUxbwos66sfcCmCH7iZHbN7q828=
=RSyl
-----END PGP PRIVATE KEY BLOCK-----
"""


@pytest.fixture()
def gnupg_key_fingerprint(gnupg_instance: GPG, private_gpg_key: str):
    keys_imported = gnupg_instance.import_keys(private_gpg_key)
    key_fingerprint = keys_imported.fingerprints[0]
    yield key_fingerprint

    if key_fingerprint in gnupg_instance.list_keys(secret=True).fingerprints:
        remove_gpg_key_pair(
            gpg_binary=gnupg_instance.gpgbinary,
            fingerprint=key_fingerprint,
        )


@pytest.fixture()
def upstream_without_config(tmp_path):
    u_remote = tmp_path / "upstream_remote"
    u_remote.mkdir()
    create_new_repo(u_remote, ["--bare"])

    return u_remote


@pytest.fixture
def create_archive():
    def create_archive_factory(output_path):
        out = io.BytesIO()
        with tarfile.open(fileobj=out, mode="w:gz") as tar:
            t = tarfile.TarInfo("dir1")
            t.type = tarfile.DIRTYPE
            tar.addfile(t)

        out.seek(0)
        with open(output_path, "bw") as archive:
            archive.write(out.read())

    return create_archive_factory



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT
import shutil
import tempfile
import unittest
from pathlib import Path

from flexmock import flexmock
from requre.modules_decorate_all_methods import (
    record_git_module,
    record_tempfile_module,
)

from packit.utils.repo import RepositoryCache

TEST_PROJECT_URL_TO_CLONE = "https://src.fedoraproject.org/rpms/python-requre.git"
TEST_PROJECT_NAME = "python-requre"


@record_tempfile_module()
@record_git_module()
class RepositoryCacheTest(unittest.TestCase):
    def setUp(self) -> None:
        self.tmp_path = Path(tempfile.mkdtemp())

    def tearDown(self) -> None:
        shutil.rmtree(self.tmp_path)

    def test_repository_cache_add_new_and_use_it(self):
        cache_path = self.tmp_path / "cache"
        clone1_path = self.tmp_path / "clone1"
        clone2_path = self.tmp_path / "clone2"
        flexmock(RepositoryCache).should_call("_clone").with_args(
            url=TEST_PROJECT_URL_TO_CLONE,
            to_path=str(cache_path / TEST_PROJECT_NAME),
            tags=True,
        )
        flexmock(RepositoryCache).should_call("_clone").with_args(
            url=TEST_PROJECT_URL_TO_CLONE,
            to_path=str(clone1_path),
            reference=str(cache_path / TEST_PROJECT_NAME),
            tags=True,
        )
        flexmock(RepositoryCache).should_call("_clone").with_args(
            url=TEST_PROJECT_URL_TO_CLONE,
            to_path=str(clone2_path),
            reference=str(cache_path / TEST_PROJECT_NAME),
            tags=True,
        )

        repo_cache = RepositoryCache(cache_path=cache_path, add_new=True)
        assert repo_cache.cached_projects == []
        assert repo_cache.projects_added == []
        assert repo_cache.projects_cloned_using_cache == []

        assert repo_cache.get_repo(url=TEST_PROJECT_URL_TO_CLONE, directory=clone1_path)
        assert repo_cache.cached_projects == [TEST_PROJECT_NAME]
        assert repo_cache.projects_added == [TEST_PROJECT_NAME]
        assert repo_cache.projects_cloned_using_cache == [TEST_PROJECT_NAME]

        assert repo_cache.get_repo(url=TEST_PROJECT_URL_TO_CLONE, directory=clone2_path)
        assert repo_cache.cached_projects == [TEST_PROJECT_NAME]
        assert repo_cache.projects_added == [TEST_PROJECT_NAME]
        assert repo_cache.projects_cloned_using_cache == [
            TEST_PROJECT_NAME,
            TEST_PROJECT_NAME,
        ]

    def test_repository_cache_do_not_add_new_if_not_enabled(self):
        cache_path = self.tmp_path / "cache"
        clone_path = self.tmp_path / "clone1"
        flexmock(RepositoryCache).should_call("_clone").with_args(
            url=TEST_PROJECT_URL_TO_CLONE,
            to_path=str(clone_path),
            tags=True,
        )

        repo_cache = RepositoryCache(cache_path=cache_path, add_new=False)
        assert repo_cache.cached_projects == []
        assert repo_cache.projects_added == []
        assert repo_cache.projects_cloned_using_cache == []

        assert repo_cache.get_repo(url=TEST_PROJECT_URL_TO_CLONE, directory=clone_path)
        assert repo_cache.cached_projects == []
        assert repo_cache.projects_added == []
        assert repo_cache.projects_cloned_using_cache == []

    def test_repository_cache_accept_str(self):
        cache_path = self.tmp_path / "cache"
        clone1_path = self.tmp_path / "clone1"
        flexmock(RepositoryCache).should_call("_clone").with_args(
            url=TEST_PROJECT_URL_TO_CLONE,
            to_path=str(cache_path / TEST_PROJECT_NAME),
            tags=True,
        )
        flexmock(RepositoryCache).should_call("_clone").with_args(
            url=TEST_PROJECT_URL_TO_CLONE,
            to_path=str(clone1_path),
            reference=str(cache_path / TEST_PROJECT_NAME),
            tags=True,
        )

        repo_cache = RepositoryCache(cache_path=str(cache_path), add_new=True)
        assert repo_cache.cached_projects == []

        assert repo_cache.get_repo(url=TEST_PROJECT_URL_TO_CLONE, directory=clone1_path)
        assert repo_cache.cached_projects == [TEST_PROJECT_NAME]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import pytest
from copr.v3 import (
    BuildProxy,
    Client,
    CoprAuthException,
    ProjectProxy,
)
from flexmock import flexmock
from munch import munchify

import packit
from packit.api import PackitAPI
from packit.config import PackageConfig
from packit.copr_helper import CoprHelper
from packit.exceptions import PackitCoprException, PackitCoprSettingsException
from tests.spellbook import run_packit


def test_copr_build_existing_project(cwd_upstream_or_distgit, api_instance):
    u, d, api = api_instance
    owner = "the-owner"
    project = "project-name"
    description = "some description"
    instructions = "the instructions"
    chroots = ["fedora-rawhide-x86_64"]

    flexmock(ProjectProxy).should_receive("add").and_return(
        flexmock(
            chroot_repos=flexmock(keys=lambda: chroots),
            description=description,
            instructions=instructions,
            unlisted_on_hp=True,
            delete_after_days=60,
            additional_repos=[],
            ownername=owner,
            module_hotfixes=False,
        ),
    )

    # no change in settings => no edit
    flexmock(ProjectProxy).should_receive("edit").times(0)

    flexmock(CoprHelper).should_receive("get_copr_client").and_return(
        Client(config={"copr_url": "https://copr.fedorainfracloud.org"}),
    )

    build = flexmock(
        id="1",
        ownername=owner,
        projectname=project,
    )
    flexmock(BuildProxy).should_receive("create_from_file").and_return(build)
    build_id, url = api.run_copr_build(
        project=project,
        owner=owner,
        chroots=chroots,
        description=description,
        instructions=instructions,
    )

    assert build_id == "1"
    assert url == f"https://copr.fedorainfracloud.org/coprs/build/{build.id}/"


def test_copr_build_existing_project_change_settings(
    cwd_upstream_or_distgit,
    api_instance,
):
    u, d, api = api_instance
    owner = "the-owner"
    project = "project-name"
    description = "some description"
    instructions = "the instructions"
    chroots = ["fedora-rawhide-x86_64"]

    flexmock(ProjectProxy).should_receive("add").and_return(
        flexmock(
            chroot_repos=flexmock(keys=lambda: chroots),
            description=description,
            instructions=instructions,
            unlisted_on_hp=True,
            delete_after_days=60,
            additional_repos=[],
            ownername=owner,
            module_hotfixes=False,
        ),
    )

    flexmock(ProjectProxy).should_receive(
        "edit",
        # ).with_args(
        # ownername="the-owner",
        # projectname="project-name",
        # chroots=["fedora-rawhide-x86_64"],
        # description="different description",
        # instructions="the instructions",
        # unlisted_on_hp=True,
        # additional_repos=None,
        # delete_after_days=60,
        #
        # Does not work:
        # flexmock.MethodSignatureError:
        # edit(
        #    <copr.v3.proxies.project.ProjectProxy object at 0x7fa53af2f3d0>,
        #    ownername="the-owner",
        #    projectname="project-name",
        #    chroots=["fedora-rawhide-x86_64"],
        #    description="different description",
        #    instructions="the instructions",
        #    unlisted_on_hp=True,
        #    additional_repos=None,
        #    delete_after_days=60,
        # )
    ).and_return().once()

    flexmock(CoprHelper).should_receive("get_copr_client").and_return(
        Client(config={"copr_url": "https://copr.fedorainfracloud.org"}),
    )

    build = flexmock(
        id="1",
        ownername=owner,
        projectname=project,
    )
    flexmock(BuildProxy).should_receive("create_from_file").and_return(build)
    build_id, url = api.run_copr_build(
        project=project,
        chroots=chroots,
        owner=owner,
        description="different description",
        instructions=instructions,
    )

    assert build_id == "1"
    assert url == f"https://copr.fedorainfracloud.org/coprs/build/{build.id}/"


def test_copr_build_existing_project_munch_no_settings_change(
    cwd_upstream_or_distgit,
    api_instance,
):
    u, d, api = api_instance
    owner = "the-owner"
    project = "project-name"
    chroots = ["fedora-rawhide-x86_64"]

    flexmock(ProjectProxy).should_receive("add").and_return(
        munchify(
            {
                "additional_repos": [],
                "auto_prune": True,
                "chroot_repos": {
                    "fedora-rawhide-x86_64": "https://download.copr.fedorainfracloud.org/"
                    "results/packit/packit-hello-world-127-stg/fedora-rawhide-x86_64/",
                },
                "contact": "https://github.com/packit/packit/issues",
                "description": "Continuous builds initiated by packit service.\n"
                "For more info check out https://packit.dev/",
                "devel_mode": False,
                "enable_net": False,
                "full_name": "packit/packit-hello-world-127-stg",
                "homepage": "",
                "id": 34245,
                "ownername": owner,
                "module_hotfixes": False,
            },
        ),
    )

    flexmock(ProjectProxy).should_receive("edit").and_return().times(0)

    flexmock(CoprHelper).should_receive("get_copr_client").and_return(
        Client(config={"copr_url": "https://copr.fedorainfracloud.org"}),
    )

    build = flexmock(
        id="1",
        ownername=owner,
        projectname=project,
    )
    flexmock(BuildProxy).should_receive("create_from_file").and_return(build)
    build_id, url = api.run_copr_build(
        project=project,
        chroots=chroots,
        owner=owner,
        description=None,
        instructions=None,
        list_on_homepage=False,
        preserve_project=False,
        additional_repos=None,
    )

    assert build_id == "1"
    assert url == f"https://copr.fedorainfracloud.org/coprs/build/{build.id}/"


def test_copr_build_existing_project_munch_additional_repos_change(
    cwd_upstream_or_distgit,
    api_instance,
):
    u, d, api = api_instance
    owner = "the-owner"
    project = "project-name"
    chroots = ["fedora-rawhide-x86_64"]

    flexmock(ProjectProxy).should_receive("add").and_return(
        munchify(
            {
                "additional_repos": [],
                "auto_prune": True,
                "chroot_repos": {
                    "fedora-rawhide-x86_64": "https://download.copr.fedorainfracloud.org/"
                    "results/packit/packit-hello-world-127-stg/fedora-rawhide-x86_64/",
                },
                "contact": "https://github.com/packit/packit/issues",
                "description": "Continuous builds initiated by packit service.\n"
                "For more info check out https://packit.dev/",
                "devel_mode": False,
                "enable_net": False,
                "full_name": "packit/packit-hello-world-127-stg",
                "homepage": "",
                "id": 34245,
                "ownername": owner,
                "module_hotfixes": False,
            },
        ),
    )

    flexmock(ProjectProxy).should_receive("edit").and_return().once()

    flexmock(CoprHelper).should_receive("get_copr_client").and_return(
        Client(config={"copr_url": "https://copr.fedorainfracloud.org"}),
    )

    build = flexmock(
        id="1",
        ownername=owner,
        projectname=project,
    )
    flexmock(BuildProxy).should_receive("create_from_file").and_return(build)
    build_id, url = api.run_copr_build(
        project=project,
        chroots=chroots,
        owner=owner,
        description=None,
        instructions=None,
        list_on_homepage=False,
        preserve_project=False,
        additional_repos=["new-repo"],
    )

    assert build_id == "1"
    assert url == f"https://copr.fedorainfracloud.org/coprs/build/{build.id}/"


def test_copr_build_existing_project_munch_list_on_homepage_change(
    cwd_upstream_or_distgit,
    api_instance,
):
    """
    We don't get that value from Copr. => We can't check the change. => No edit.
    """

    u, d, api = api_instance
    owner = "the-owner"
    project = "project-name"
    chroots = ["fedora-rawhide-x86_64"]

    flexmock(ProjectProxy).should_receive("add").and_return(
        munchify(
            {
                "additional_repos": [],
                "auto_prune": True,
                "chroot_repos": {
                    "fedora-rawhide-x86_64": "https://download.copr.fedorainfracloud.org/"
                    "results/packit/packit-hello-world-127-stg/fedora-rawhide-x86_64/",
                },
                "contact": "https://github.com/packit/packit/issues",
                "description": "Continuous builds initiated by packit service.\n"
                "For more info check out https://packit.dev/",
                "devel_mode": False,
                "enable_net": False,
                "full_name": "packit/packit-hello-world-127-stg",
                "homepage": "",
                "id": 34245,
                "ownername": owner,
                "module_hotfixes": False,
            },
        ),
    )

    # We don't get that value from Copr. => We can't check the change. => No edit.
    flexmock(ProjectProxy).should_receive("edit").and_return().times(0)

    flexmock(CoprHelper).should_receive("get_copr_client").and_return(
        Client(config={"copr_url": "https://copr.fedorainfracloud.org"}),
    )

    build = flexmock(
        id="1",
        ownername=owner,
        projectname=project,
    )
    flexmock(BuildProxy).should_receive("create_from_file").and_return(build)
    build_id, url = api.run_copr_build(
        project=project,
        chroots=chroots,
        owner=owner,
        description=None,
        instructions=None,
        list_on_homepage=True,
        preserve_project=False,
        additional_repos=None,
    )

    assert build_id == "1"
    assert url == f"https://copr.fedorainfracloud.org/coprs/build/{build.id}/"


def test_copr_build_existing_project_munch_do_not_update_booleans_by_default(
    cwd_upstream_or_distgit,
    api_instance,
):
    u, d, api = api_instance
    owner = "the-owner"
    project = "project-name"
    chroots = ["fedora-rawhide-x86_64"]

    flexmock(ProjectProxy).should_receive("add").and_return(
        munchify(
            {
                "additional_repos": [],
                "auto_prune": True,
                "chroot_repos": {
                    "fedora-rawhide-x86_64": "https://download.copr.fedorainfracloud.org/"
                    "results/packit/packit-hello-world-127-stg/fedora-rawhide-x86_64/",
                },
                "contact": "https://github.com/packit-service/packit/issues",
                "description": "Continuous builds initiated by packit service.\n"
                "For more info check out https://packit.dev/",
                "unlisted_on_hp": True,  # Value not present currently.
                "devel_mode": False,
                "enable_net": False,
                "full_name": "packit/packit-hello-world-127-stg",
                "homepage": "",
                "id": 34245,
                "ownername": owner,
                "module_hotfixes": False,
            },
        ),
    )

    # Even if we receive this info from Copr, we can't edit that value if it is `None`.
    flexmock(ProjectProxy).should_receive("edit").and_return().times(0)

    flexmock(CoprHelper).should_receive("get_copr_client").and_return(
        Client(config={"copr_url": "https://copr.fedorainfracloud.org"}),
    )

    build = flexmock(
        id="1",
        ownername=owner,
        projectname=project,
    )
    flexmock(BuildProxy).should_receive("create_from_file").and_return(build)
    build_id, url = api.run_copr_build(
        project=project,
        chroots=chroots,
        owner=owner,
        description=None,
        instructions=None,
        list_on_homepage=None,
        preserve_project=False,
        additional_repos=None,
    )

    assert build_id == "1"
    assert url == f"https://copr.fedorainfracloud.org/coprs/build/{build.id}/"


@pytest.mark.parametrize(
    "owner, requested_chroots, expected_chroots_for_edit",
    (
        # For Copr projects that are not created by Packit we try to extend the
        # chroots instead of replacing them.
        (
            "the-owner",
            ["fedora-rawhide-x86_64", "fedora-35-x86_64"],
            ["fedora-rawhide-x86_64", "fedora-35-x86_64", "epel-8-x86_64"],
        ),
        # For Copr projects that are not created by Packit we **do not** touch
        # chroots as long as the requirements for Copr build are satisfied.
        ("the-owner", ["fedora-rawhide-x86_64"], None),
    ),
)
def test_copr_build_existing_project_munch_chroot_updates(
    cwd_upstream_or_distgit,
    api_instance,
    owner,
    requested_chroots,
    expected_chroots_for_edit,
):
    u, d, api = api_instance
    project = "project-name"

    flexmock(ProjectProxy).should_receive("add").and_return(
        munchify(
            {
                "additional_repos": [],
                "auto_prune": True,
                "chroot_repos": {
                    "fedora-rawhide-x86_64": "https://download.copr.fedorainfracloud.org/"
                    "results/packit/packit-hello-world-127-stg/fedora-rawhide-x86_64/",
                    "epel-8-x86_64": "https://download.copr.fedorainfracloud.org/"
                    "results/packit/packit-hello-world-127-stg/epel-8-x86_64/",
                },
                "contact": "https://github.com/packit-service/packit/issues",
                "description": "Continuous builds initiated by packit service.\n"
                "For more info check out https://packit.dev/",
                "unlisted_on_hp": True,  # Value not present currently.
                "devel_mode": False,
                "enable_net": False,
                "full_name": "packit/packit-hello-world-127-stg",
                "homepage": "",
                "id": 34245,
                "ownername": owner,
                "module_hotfixes": False,
            },
        ),
    )

    if expected_chroots_for_edit:
        expected_chroots_for_edit.sort()
        flexmock(ProjectProxy).should_receive("edit").with_args(
            ownername=owner,
            projectname=project,
            chroots=expected_chroots_for_edit,
        ).and_return().once()
    else:
        flexmock(ProjectProxy).should_receive("edit").never()

    flexmock(CoprHelper).should_receive("get_copr_client").and_return(
        Client(config={"copr_url": "https://copr.fedorainfracloud.org"}),
    )

    build = flexmock(
        id="1",
        ownername=owner,
        projectname=project,
    )
    flexmock(BuildProxy).should_receive("create_from_file").and_return(build)
    build_id, url = api.run_copr_build(
        project=project,
        chroots=requested_chroots,
        owner=owner,
        description=None,
        instructions=None,
        list_on_homepage=None,
        preserve_project=False,
        additional_repos=None,
    )

    assert build_id == "1"
    assert url == f"https://copr.fedorainfracloud.org/coprs/build/{build.id}/"


def test_copr_build_existing_project_error_on_change_settings(
    cwd_upstream_or_distgit,
    api_instance,
):
    u, d, api = api_instance
    owner = "the-owner"
    project = "project-name"
    description = "some description"
    instructions = "the instructions"
    chroots = ["fedora-rawhide-x86_64"]

    flexmock(ProjectProxy).should_receive("add").and_return(
        flexmock(
            chroot_repos=flexmock(keys=lambda: chroots),
            description=description,
            instructions=instructions,
            unlisted_on_hp=True,
            delete_after_days=60,
            additional_repos=[],
            ownername=owner,
            module_hotfixes=False,
        ),
    )

    flexmock(ProjectProxy).should_receive("request_permissions").with_args(
        ownername=owner,
        projectname=project,
        permissions={"admin": True},
    ).and_return()

    flexmock(ProjectProxy).should_receive("edit").and_raise(
        CoprAuthException,
        "Only owners and admins may update their projects.",
    ).once()

    flexmock(CoprHelper).should_receive("get_copr_client").and_return(
        Client(config={"copr_url": "https://copr.fedorainfracloud.org"}),
    )

    build = flexmock(
        id="1",
        ownername=owner,
        projectname=project,
    )
    flexmock(BuildProxy).should_receive("create_from_file").and_return(build)

    with pytest.raises(PackitCoprSettingsException) as e_info:
        api.run_copr_build(
            project=project,
            chroots=chroots,
            owner=owner,
            description="different description",
            instructions=instructions,
        )
    assert e_info.value.fields_to_change == {
        "description": ("some description", "different description"),
    }


def test_copr_build_no_owner(cwd_upstream_or_distgit, api_instance):
    u, d, api = api_instance
    flexmock(CoprHelper).should_receive("get_copr_client").and_return(
        Client(config={"copr_url": "https://copr.fedorainfracloud.org"}),
    )
    with pytest.raises(PackitCoprException) as ex:
        api.run_copr_build(
            project="project-name",
            chroots="fedora-rawhide-x86_64",
            owner=None,
            description="some description",
            instructions="the instructions",
        )
    assert "owner not set" in str(ex)


def test_copr_build_cli_no_project_configured(upstream_and_remote, copr_client_mock):
    upstream, _ = upstream_and_remote
    flexmock(PackitAPI).should_receive("run_copr_build").with_args(
        project="packit-cli-upstream_remote-upstream_git-main",
        chroots=["fedora-rawhide-x86_64"],
        owner=None,
        description=None,
        instructions=None,
        upstream_ref=None,
        list_on_homepage=False,
        preserve_project=False,
        additional_repos=None,
        bootstrap=None,
        request_admin_if_needed=False,
        enable_net=False,
        release_suffix=None,
        srpm_path=None,
        module_hotfixes=False,
    ).and_return(("id", "url")).once()

    flexmock(packit.copr_helper.CoprClient).should_receive(
        "create_from_config_file",
    ).and_return(copr_client_mock)
    CoprHelper.get_available_chroots.cache_clear()

    run_packit(["build", "in-copr", "--no-wait"], working_dir=upstream)


def test_copr_build_cli_project_set_via_cli(upstream_and_remote, copr_client_mock):
    upstream, _ = upstream_and_remote
    flexmock(PackitAPI).should_receive("run_copr_build").with_args(
        project="the-project",
        chroots=["fedora-rawhide-x86_64"],
        owner=None,
        description=None,
        instructions=None,
        upstream_ref=None,
        list_on_homepage=False,
        preserve_project=False,
        additional_repos=None,
        bootstrap=None,
        request_admin_if_needed=False,
        enable_net=False,
        release_suffix=None,
        srpm_path=None,
        module_hotfixes=False,
    ).and_return(("id", "url")).once()

    flexmock(packit.copr_helper.CoprClient).should_receive(
        "create_from_config_file",
    ).and_return(copr_client_mock)
    CoprHelper.get_available_chroots.cache_clear()

    run_packit(
        ["build", "in-copr", "--no-wait", "--project", "the-project"],
        working_dir=upstream,
    )


def test_copr_build_cli_project_set_from_config(upstream_and_remote, copr_client_mock):
    upstream, _ = upstream_and_remote

    flexmock(PackageConfig).should_receive("get_copr_build_project_value").and_return(
        "some-project",
    )
    flexmock(packit.copr_helper.CoprClient).should_receive(
        "create_from_config_file",
    ).and_return(copr_client_mock)
    CoprHelper.get_available_chroots.cache_clear()

    flexmock(PackitAPI).should_receive("run_copr_build").with_args(
        project="some-project",
        chroots=["fedora-rawhide-x86_64"],
        owner=None,
        description=None,
        instructions=None,
        upstream_ref=None,
        list_on_homepage=False,
        preserve_project=False,
        additional_repos=None,
        bootstrap=None,
        request_admin_if_needed=False,
        enable_net=False,
        release_suffix=None,
        srpm_path=None,
        module_hotfixes=False,
    ).and_return(("id", "url")).once()

    run_packit(["build", "in-copr", "--no-wait"], working_dir=upstream)


def test_create_or_update_copr_project(copr_client_mock):
    copr_helper = CoprHelper(flexmock(git_url="https://gitlab.com/"))
    flexmock(packit.copr_helper.CoprClient).should_receive(
        "create_from_config_file",
    ).and_return(copr_client_mock)

    options = {
        "chroots": ["centos-stream-8-x86_64"],
        "description": "my fabulous test",
        "instructions": None,
        "owner": "me",
        "project": "already-present",
        "targets_dict": {"centos-stream-8": {"additional_packages": ["foo"]}},
        "module_hotfixes": None,
    }

    copr_client_mock.project_proxy = flexmock()
    copr_client_mock.project_chroot_proxy = flexmock()
    flexmock(copr_client_mock.project_proxy).should_receive("add").and_return(
        flexmock(
            chroot_repos={"centos-stream-8-x86_64": "https://repo.url"},
            **options,
        ),
    )
    flexmock(copr_client_mock.project_chroot_proxy).should_receive("get").and_return(
        {"additional_packages": []},
    )
    flexmock(copr_client_mock.project_chroot_proxy).should_receive("edit").with_args(
        ownername="me",
        projectname="already-present",
        chrootname="centos-stream-8-x86_64",
        additional_packages=["foo"],
    ).and_return({})

    copr_helper.create_or_update_copr_project(**options)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from munch import Munch

BODHI_LATEST_BUILDS = Munch(
    {
        "f28-updates-candidate": "colin-0.3.1-1.fc28",
        "f29-updates-candidate": "colin-0.3.1-1.fc29",
        "f30-updates-candidate": "colin-0.3.1-2.fc30",
        "f27-updates-candidate": "colin-0.2.0-1.fc27",
        "f28-updates-testing": "colin-0.3.1-1.fc28",
        "f29-updates-testing": "colin-0.3.1-1.fc29",
        "f27-updates-testing": "colin-0.2.0-1.fc27",
        "f30-updates-testing": "colin-0.3.1-2.fc30",
        "f28-updates": "colin-0.3.1-1.fc28",
        "f29-updates": "colin-0.3.1-1.fc29",
        "f27-updates": "colin-0.2.0-1.fc27",
        "f30-updates": "colin-0.3.1-2.fc30",
        "f28-override": "colin-0.3.1-1.fc28",
        "f29-override": "colin-0.3.1-1.fc29",
        "f27-override": "colin-0.2.0-1.fc27",
        "f30-override": "colin-0.3.1-2.fc30",
        "f28-updates-testing-pending": "colin-0.3.1-1.fc28",
        "f29-updates-testing-pending": "colin-0.3.1-1.fc29",
        "f27-updates-testing-pending": "colin-0.0.4-3.fc27",
        "f30-updates-testing-pending": "colin-0.3.1-2.fc30",
        "f28-updates-pending": "colin-0.3.1-1.fc28",
        "f29-updates-pending": "colin-0.3.1-1.fc29",
        "f27-updates-pending": "colin-0.2.0-1.fc27",
        "f30-updates-pending": "colin-0.3.1-2.fc30",
    },
)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

from bodhi.client.oidcclient import OIDCClient
from flexmock import flexmock
from requests import Session

from packit.utils.bodhi import get_bodhi_client


def test_login_with_kerberos():
    bodhi = get_bodhi_client()
    bodhi.oidc._tokens = {"access_token": "token"}
    the_code = (
        "code=d37deb2e-5463-1234-1234-ba70c75d74f4_5EH4MhV3Lj2ld1HapvvMM_r4Vy-eFX6R"
        "&amp;state=k44Rw12G86v6605oTNv6mm7yhZRGXO"
    )
    flexmock(Session).should_receive("request").and_return(
        flexmock(
            raise_for_status=lambda: None,
            text=f'<meta charset="UTF-8">\n<title>{the_code}</title>\n   ',
        ),
    )
    flexmock(OIDCClient).should_receive("auth_callback").with_args(f"?{the_code}")
    bodhi.oidc.login_with_kerberos("foobar")



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
This module contains code related to security, signing and verification.
"""

import logging
from enum import Enum
from typing import Optional

import git
from gnupg import GPG, ListKeys

from packit.exceptions import PackitException

logger = logging.getLogger(__name__)


class CommitVerifier:
    """
    Class used for verifying git commits. Uses python-gnupg for accessing the GPG binary.
    """

    def __init__(self, key_server: Optional[str] = None) -> None:
        """
        :param key_server: GPG key server to be used
        """
        self._gpg: Optional[GPG] = None
        if key_server:
            self.key_servers = [key_server]
        else:
            self.key_servers = [
                "keyserver.ubuntu.com",
                "keys.openpgp.org",
                "pgp.surf.nl",
                # "pgp.mit.edu",  # slow & unreliable
            ]

    @property
    def gpg(self) -> GPG:
        """
        gnupg.GPG instance from python-gnupg
        """
        if not self._gpg:
            self._gpg = GPG()
        return self._gpg

    @property
    def _gpg_keys(self) -> ListKeys:
        """
        List of installed gpg keys

        (do not cache this value)
        """
        return self.gpg.list_keys()

    @property
    def _gpg_fingerprints(self) -> list[str]:
        """List of fingerprints of the saved keys"""
        return self._gpg_keys.fingerprints

    def download_gpg_key_if_needed(self, key_fingerprint: str) -> str:
        """
        Download the gpg key from the self.key_servers if it is not present.

        :param key_fingerprint: fingerprint of the gpg key
        """
        if key_fingerprint in self._gpg_fingerprints:
            return key_fingerprint

        try:
            for keyserver in self.key_servers:
                logger.debug(f"Downloading {key_fingerprint!r} from {keyserver!r}.")
                result = self.gpg.recv_keys(keyserver, key_fingerprint)
                if result.fingerprints:
                    return result.fingerprints[0]
        except Exception as ex:
            raise PackitException(
                f"Cannot receive a gpg key: {key_fingerprint}",
            ) from ex

        raise PackitException(f"Cannot receive a gpg key: {key_fingerprint}")

    def check_signature_of_commit(
        self,
        commit: git.Commit,
        possible_key_fingerprints: list[str],
    ) -> bool:
        """
        Check the validity of the commit signature
        and test if the signer is present in the provided list.
        (Commit without signature returns False.)
        """
        status = self.get_commit_signature_status(commit=commit)
        if status == CommitSignatureStatus.no_signature:
            logger.debug("Commit not signed.")
            return False

        if status == CommitSignatureStatus.cannot_be_checked:
            # We need to download keys before getting the signer
            for key in possible_key_fingerprints:
                self.download_gpg_key_if_needed(key_fingerprint=key)

        signer = self.get_commit_signer_fingerprint(commit)

        if not signer:
            logger.debug("Cannot get a signer of the commit.")
            return False

        if signer not in possible_key_fingerprints:
            logger.warning("Signature author not authorized.")
            return False

        is_valid = self.is_commit_signature_valid(commit)
        if not is_valid:
            logger.warning(f"Commit {commit.hexsha!r} signature is not valid.")
        return is_valid

    def is_commit_signature_valid(self, commit: git.Commit) -> bool:
        """
        Check the validity of the commit signature.
        Key needs to be already present.
        """
        commit_status = self.get_commit_signature_status(commit)
        if commit_status in VALID_SIGNATURE_STATUSES:
            logger.debug(f"Commit {commit.hexsha!r} signature is valid.")
            return True

        logger.warning(f"Commit {commit.hexsha!r} signature is not valid.")
        return False

    @staticmethod
    def get_commit_signature_status(commit: git.Commit) -> "CommitSignatureStatus":
        """Get a signature status from the given commit."""
        signature_mark = CommitVerifier._get_commit_info(commit, pretty_format="%G?")
        return CommitSignatureStatus(signature_mark)

    @staticmethod
    def get_commit_signer_fingerprint(commit: git.Commit) -> str:
        """Get a signer fingerprint from the given commit"""
        return CommitVerifier._get_commit_info(commit, pretty_format="%GF")

    @staticmethod
    def _get_commit_info(commit: git.Commit, pretty_format: str) -> str:
        """
        Return a commit information in a given format.

        See `git show --help` and `--pretty=format` for more information.
        """
        try:
            return commit.repo.git.show(commit.hexsha, pretty=f"format:{pretty_format}")
        except git.GitCommandError as error:
            raise PackitException(
                f"Cannot find commit {commit.hexsha!r} to check its signature.",
            ) from error


class CommitSignatureStatus(Enum):
    no_signature = "N"
    bad = "B"
    good_valid = "G"
    good_unknown_validity = "U"
    good_expired = "X"
    good_made_by_expired_key = "Y"
    good_made_by_revoked_key = "R"
    cannot_be_checked = "E"


VALID_SIGNATURE_STATUSES = [
    CommitSignatureStatus.good_valid,
    CommitSignatureStatus.good_unknown_validity,
]



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Push Bodhi updates from testing to stable.
"""

import logging
import os

import click

from packit.cli.types import LocalProjectParameter
from packit.cli.utils import cover_packit_exception, get_packit_api
from packit.config import get_context_settings, pass_config

logger = logging.getLogger(__name__)


@click.command("push-updates", context_settings=get_context_settings())
@click.option("--update-alias", help="For example FEDORA-2019-ee5674e22c", default=None)
@click.argument("path_or_url", type=LocalProjectParameter(), default=os.path.curdir)
@pass_config
@cover_packit_exception
def push_updates(update_alias, config, path_or_url):
    """
    Find all Bodhi updates that have been in testing for more than 'Stable days' (7 by default)
    and push them to stable.

    If you are not authenticated with the bodhi server, please make sure that you
    navigate in your browser to the URL provided by the bodhi-client and then paste
    the `code=XX...` to the terminal when prompted.

    If you set `fas_user` and `kerberos_realm` in your "~/.config/packit.yaml" and
    have an active Kerberos TGT, you will be automatically authenticated. Otherwise,
    you need to follow the prompt
    """
    api = get_packit_api(config=config, local_project=path_or_url)
    api.push_updates(update_alias)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

import shutil
import textwrap
from pathlib import Path

import git
import pytest
from flexmock import flexmock

from packit.exceptions import PackitException
from packit.patches import PatchGenerator, PatchMetadata
from tests.spellbook import DATA_DIR

TESTS_DIR = str(Path(__file__).parent.parent)


def check_copytree_dirs_exists_support():
    """
    Old version of shutil.copytree does not support
    dirs_exist_ok parameter.
    """
    return bool(
        hasattr(shutil.copytree, "__defaults__")
        and len(shutil.copytree.__defaults__) >= 5,
    )


@pytest.fixture
def git_repo(tmp_path: Path) -> git.Repo:
    """
    Set up a git repo with some initial patch files in the history,
    and the same patch files updated after re-generating them from a
    source-git repo.
    """
    repo_dir = tmp_path.joinpath("repo")
    shutil.copytree(
        src=f"{DATA_DIR}/patches/previous/",
        dst=str(repo_dir),
    )
    repo = git.Repo.init(repo_dir)
    repo.git.add(repo.working_tree_dir)
    repo.git.commit("-mInitial patches")
    shutil.copytree(
        src=f"{DATA_DIR}/patches/regenerated/",
        dst=repo.working_tree_dir,
        dirs_exist_ok=True,
    )
    return repo


@pytest.mark.skipif(
    not check_copytree_dirs_exists_support(),
    reason="Old python version does not support copytree exists dirs parameter"
    " https://github.com/packit/packit/issues/1160",
)
def test_undo_identical(git_repo):
    """
    Check that identical patches are correctly detected and changes
    undone in the target git repo.
    """
    input_patch_list = [
        PatchMetadata(name=path.name, path=path)
        for path in Path(git_repo.working_tree_dir).iterdir()
        if path.suffix == ".patch"
    ]
    output_patch_list = [
        x for x in input_patch_list if x.name == "weird-identical.patch"
    ]
    assert (
        PatchGenerator.undo_identical(input_patch_list, git_repo) == output_patch_list
    )
    # 'weird-identical.patch' is identical, except the original patch file
    # is missing a "function" name at one of the hunks, which causes the
    # patch-ids to be different.
    # Is there any safe way to handle this?
    assert [item.a_path for item in git_repo.index.diff(None)] == [
        "weird-identical.patch",
    ]


@pytest.fixture
def source_git_repo(tmp_path: Path) -> git.Repo:
    repo_dir = tmp_path.joinpath("src/hello")
    shutil.copytree(
        src=f"{DATA_DIR}/patches/src/hello/",
        dst=str(repo_dir),
    )
    repo = git.Repo.init(repo_dir)
    repo.git.add(repo.working_dir)
    repo.git.commit("-mInitial commit")
    repo.git.tag("0.1.0")
    return repo


@pytest.fixture
def dist_git_repo(tmp_path: Path) -> git.Repo:
    repo_dir = tmp_path.joinpath("rpms/hello")
    shutil.copytree(
        src=f"{DATA_DIR}/patches/rpms/hello/",
        dst=str(repo_dir),
    )
    repo = git.Repo.init(repo_dir)
    repo.git.add(repo.working_dir)
    repo.git.commit("-mInitial commit")
    return repo


def create_commits_to_squash(repo: git.Repo):
    readme = Path(repo.working_dir, "README.md")

    readme.write_text(f"{readme.read_text()}\nThe first standalone patch.\n")
    repo.git.add("README.md")
    repo.git.commit("-mAdd a standalone patch")

    readme.write_text(f"{readme.read_text()}\nThe first commit of the second patch.\n")
    repo.git.add("README.md")
    commit_message = textwrap.dedent(
        """\
        Add a distro change

        patch_name: distro.patch
        """,
    )
    repo.git.commit(f"-m{commit_message}")

    readme.write_text(f"{readme.read_text()}\nThe second commit of the second patch.\n")
    repo.git.add("README.md")
    commit_message = textwrap.dedent(
        """\
        Another distro change

        patch_name: distro.patch
        """,
    )
    repo.git.commit(f"-m{commit_message}")

    readme.write_text(f"{readme.read_text()}\nThe second standalone patch.\n")
    repo.git.add("README.md")
    repo.git.commit("-mAdd another standalone patch")


def test_squash_patches_by_name(source_git_repo: git.Repo, dist_git_repo: git.Repo):
    """Patch files corresponding to commits which have identical 'patch_name'
    metadata defined are squashed.
    """
    local_project = flexmock(
        git_repo=source_git_repo,
        ref="HEAD",
        working_dir=source_git_repo.working_dir,
    )

    create_commits_to_squash(source_git_repo)

    patch_generator = PatchGenerator(local_project)
    patch_list = patch_generator.create_patches(
        git_ref="0.1.0",
        destination=dist_git_repo.working_dir,
    )
    assert len(patch_list) == 3
    assert patch_list[1].path == Path(dist_git_repo.working_dir, "distro.patch")
    assert patch_list[1].name == "distro.patch"
    assert sorted(dist_git_repo.untracked_files) == [
        "0001-Add-a-standalone-patch.patch",
        "0004-Add-another-standalone-patch.patch",
        "distro.patch",
    ]
    patch = Path(dist_git_repo.working_dir, "distro.patch").read_text()
    assert "+The first commit of the second patch." in patch
    assert "+The second commit of the second patch." in patch


def create_non_adjacent_commits_to_squash(repo: git.Repo):
    """Patch files corresponding to non-adjacent commits, which should be squashed
    by name are not squashed and an exception is raised."""
    readme = Path(repo.working_dir, "README.md")

    readme.write_text(f"{readme.read_text()}\nThe first change.\n")
    repo.git.add("README.md")
    commit_message = textwrap.dedent(
        """\
        First commit

        patch_name: non_adjacent.patch
        """,
    )
    repo.git.commit(f"-m{commit_message}")

    readme = Path(repo.working_dir, "README.md")
    readme.write_text(f"{readme.read_text()}\nThe second change.\n")
    repo.git.add("README.md")
    commit_message = "Second commit"
    repo.git.commit(f"-m{commit_message}")

    readme = Path(repo.working_dir, "README.md")
    readme.write_text(f"{readme.read_text()}\nThe third change.\n")
    repo.git.add("README.md")
    commit_message = textwrap.dedent(
        """\
        Third commit

        patch_name: non_adjacent.patch
        """,
    )
    repo.git.commit(f"-m{commit_message}")


def test_fail_if_not_adjacent(source_git_repo: git.Repo, dist_git_repo: git.Repo):
    local_project = flexmock(
        git_repo=source_git_repo,
        ref="HEAD",
        working_dir=source_git_repo.working_dir,
    )

    create_non_adjacent_commits_to_squash(source_git_repo)

    patch_generator = PatchGenerator(local_project)
    with pytest.raises(PackitException) as ex:
        patch_generator.create_patches(
            git_ref="0.1.0",
            destination=dist_git_repo.working_dir,
        )
    assert "Non-adjacent patches" in str(ex)



# Copyright Contributors to the Packit project.
# SPDX-License-Identifier: MIT

"""
Processing RPM spec file patches.
"""
import datetime
import email
import logging
import re
import tempfile
from itertools import islice
from pathlib import Path
from typing import Optional

import git
import yaml

from packit.constants import DATETIME_FORMAT, PATCH_META_TRAILER_TOKENS
from packit.exceptions import PackitException, PackitGitException
from packit.local_project import LocalProject
from packit.utils.commands import run_command
from packit.utils.repo import get_metadata_from_message, git_patch_ish, is_a_git_ref

logger = logging.getLogger(__name__)


def git_format_patch(
    working_dir: Path,
    destination: str,
    files_to_ignore: list[str],
    ref_or_range: str,
    no_prefix: bool = False,
):
    """
    run `git format-patch $ref_or_range` in working_dir

    :param working_dir: Path where to run the command.
    :param destination: place the patches here
    :param files_to_ignore: ignore changes in these files
    :param ref_or_range: [ <since> | <revision range> ]:
           1. A single commit, <since>, specifies that the commits leading to the tip of the
              current branch that are not in the history that leads to the <since> to be
              output.

           2. Generic <revision range> expression (see "SPECIFYING REVISIONS" section in
              gitrevisions(7)) means the commits in the specified range.
    :param no_prefix: prefix is the leading a/ and b/ - format-patch does this by default
    :return: str, git format-patch output: new-line separated list of patch names
    """
    git_f_p_cmd = ["git", "format-patch", "--output-directory", f"{destination}"]
    if no_prefix:
        git_f_p_cmd.append("--no-prefix")
    git_f_p_cmd += [
        ref_or_range,
        "--",
        ".",
    ] + [f":(exclude){file_to_ignore}" for file_to_ignore in files_to_ignore]
    return run_command(
        cmd=git_f_p_cmd,
        cwd=working_dir,
        output=True,
    ).stdout.strip()


def git_interpret_trailers(patch: str) -> str:
    """Run 'git interpret-trailers' on 'patch' and return the output.

    We don't use --parse/--unfold to unfold multiline values per RFC822
    because we use YAML block scalars for multiline trailer values.

    Args:
        patch: Path of the patch.

    Returns:
        The output of the command.
    """
    cmd = ["git", "interpret-trailers", "--only-input", "--only-trailers", patch]
    return run_command(cmd=cmd, cwd=Path.cwd(), output=True).stdout


def remove_prefixes(patch: Path):
    """Remove source or destination prefixes from a patch file

    Args:
        patch: Path of the patch file.
    """
    logger.debug(f"Remove prefixes from {patch}.")
    content = patch.read_text()
    content = re.sub(
        r"^(diff .+) a/(.+) b/(.+)$",
        r"\1 \2 \3",
        content,
        flags=re.MULTILINE,
    )
    content = re.sub(r"^(---) a/(.+)$", r"\1 \2", content, flags=re.MULTILINE)
    content = re.sub(r"^(\+\+\+) b/(.+)$", r"\1 \2", content, flags=re.MULTILINE)
    patch.write_text(content)


def commit_message(
    patch: Path,
    strip_subject_prefix: Optional[str] = None,
    strip_trailers: Optional[str] = None,
) -> str:
    """Read the commit message from a file

    The file can be a text file with the commit message or a patch file produced by
    'git format-patch'.

    If specified, strip any '[<strip_subject_prefix>]' prefix from the subject line
    and the git trailers.

    Args:
        patch: Path of the file.
        strip_subject_prefix: Strip this prefix from subject lines.
        strip_trailers: Strip these trailer lines from the end of the commit message.

    Returns:
        The commit message: subject and body, separated by an empty line.
    """
    message = email.message_from_bytes(patch.read_bytes())
    subject = message["Subject"]
    # The file only contains the commit message.
    if not subject:
        subject, _, body = message.get_payload().partition("\n\n")
    # The file is a patch file generated by format-patch.
    else:
        body, sep, _ = message.get_payload().partition("\n---\n")
        if not sep:
            body = ""

    if strip_subject_prefix:
        # Strip [<subject_prefix>] or [<subject_prefix> n/m]
        subject = re.sub(
            r"^\[" + strip_subject_prefix + r"( \d+/\d+)?\] (.+)$",
            r"\2",
            subject,
        )

    if strip_trailers:
        body = body.partition(strip_trailers)[0]

    empty_line = "\n\n" if body else ""
    return f"{subject}{empty_line}{body}".strip()


class PatchMetadata:
    """
    Metadata about patch files and relation to the respective commit.

    Attributes:
        name: name of the patch file (patch_name is the actual metadata key)
        path: Path of the patch file
        description: will be attached as a comment above Patch spec definition
                     (if present_in_specfile=False)
        commit: git.Commit relevant to this patch file
        present_in_specfile: if the patch is already in the spec-file
                             and we don't need to add it there
        ignore: this patch will not be processed.
        squash_commits: squash commits into a single patch file
                        until next commits with squash_commits=True
                        (git-am patches do this)
        patch_id: the number from definition PatchNNNN in the spec
        no_prefix: do not prepend a/ and b/ when generating the patch file
        metadata_defined: are any of the metadata defined in a commit message
    """

    def __init__(
        self,
        name: Optional[str] = None,
        path: Optional[Path] = None,
        description: Optional[str] = None,
        commit: Optional[git.Commit] = None,
        present_in_specfile: bool = False,
        ignore: bool = False,
        squash_commits: bool = False,
        patch_id: Optional[int] = None,
        no_prefix: bool = False,
        metadata_defined: bool = False,
    ) -> None:
        self.name = name
        self.path = path
        self.description = description
        self.commit = commit
        self.present_in_specfile = present_in_specfile
        self.ignore = ignore
        self.squash_commits = squash_commits
        self.patch_id = patch_id
        self.no_prefix = no_prefix
        # this is set during PatchMetadata.from_commit()
        self.metadata_defined = metadata_defined

    def __eq__(self, other) -> bool:
        return (
            self.name == other.name
            and self.path == other.path
            and self.description == other.description
            and self.commit == other.commit
            and self.present_in_specfile == other.present_in_specfile
            and self.ignore == other.ignore
            and self.squash_commits == other.squash_commits
            and self.patch_id == other.patch_id
            and self.no_prefix == other.no_prefix
            and self.metadata_defined == other.metadata_defined
        )

    @property
    def specfile_comment(self) -> str:
        if self.description:
            comment = self.description
        elif self.commit:
            comment = (
                f"{self.commit.summary}\n"
                f"Author: {self.commit.author.name} <{self.commit.author.email}>"
            )
        else:
            comment = f'Patch "{self.name}"'
        return comment

    @property
    def commit_message(self) -> str:
        msg = f"Apply {self.name}\n"

        if self.name:
            msg += f"\npatch_name: {self.name}"

        if self.description:
            msg += f"\ndescription: {self.description}"

        if self.present_in_specfile:
            msg += "\npresent_in_specfile: true"

        if self.ignore:
            msg += "\nignore: true"

        if self.squash_commits:
            msg += "\nsquash_commits: true"

        if self.patch_id is not None:
            msg += f"\npatch_id: {self.patch_id}"

        if self.no_prefix:
            msg += "\nno_prefix: true"

        return msg

    @staticmethod
    def from_commit(
        commit: git.Commit,
        patch_path: Optional[Path] = None,
    ) -> "PatchMetadata":
        """
        Load PatchMetadata from an existing git.Commit

        @param commit: the git.Commit to load the metadata from
        @param patch_path: optional Path to an existing patch file present on the disk
        @return: PatchMetadata instance
        """
        metadata = get_metadata_from_message(commit) or {}
        metadata_defined = False
        if metadata:
            logger.debug(
                f"Commit {commit.hexsha:.8} metadata:\n"
                f"{yaml.dump(metadata, indent=4, default_flow_style=False)}",
            )
            metadata_defined = True
        else:
            logger.debug(f"Commit {commit.hexsha:.8} does not contain any metadata.")

        name = metadata.get("patch_name") or (patch_path.name if patch_path else None)

        return PatchMetadata(
            name=name,
            path=patch_path,
            description=metadata.get("description"),
            present_in_specfile=metadata.get("present_in_specfile"),
            ignore=metadata.get("ignore"),
            commit=commit,
            squash_commits=metadata.get("squash_commits"),
            patch_id=metadata.get("patch_id", None),
            no_prefix=metadata.get("no_prefix"),
            metadata_defined=metadata_defined,
        )

    @staticmethod
    def from_git_trailers(commit: git.Commit) -> "PatchMetadata":
        """Read patch metadata from a commit's git trailers

        Args:
            commit: Commit object to read patch metadata from.

        Returns:
            Patch metadata read.
        """
        with tempfile.NamedTemporaryFile(mode="w+t") as fp:
            fp.write(commit.message)
            fp.flush()
            return PatchMetadata.from_patch(fp.name)

    @staticmethod
    def from_patch(patch: str) -> "PatchMetadata":
        """Read patch metadata by parsing Git trailers from a patch file.

        Args:
            patch: Path of a patch file.

        Returns:
            A PatchMetadata object.
        """
        trailers = git_interpret_trailers(patch).strip()
        trailer = ""
        metadata = {}
        for line in trailers.splitlines():
            token = line.partition(": ")[0]
            if token and not token.startswith(" ") and trailer:
                # found ANOTHER trailer, store the previous one
                metadata.update(yaml.safe_load(trailer))
                trailer = ""
            if token in PATCH_META_TRAILER_TOKENS or (
                token.startswith(" ") and trailer
            ):
                # it's OUR token or we're already processing OUR trailer
                trailer += line + "\n"
        metadata.update(yaml.safe_load(trailer) or {})

        patch_path = Path(patch)
        description = metadata.get("Patch-status") or commit_message(
            patch_path,
            strip_subject_prefix="PATCH",
            strip_trailers=trailers,
        )
        return PatchMetadata(
            name=metadata.get("Patch-name") or patch_path.name,
            path=patch_path,
            description=description.strip(),
            present_in_specfile=metadata.get("Patch-present-in-specfile", False),
            ignore=metadata.get("Ignore-patch", False),
            patch_id=metadata.get("Patch-id"),
            no_prefix=metadata.get("No-prefix", False),
            metadata_defined=bool(metadata),
        )

    def __repr__(self):
        return (
            f"Patch("
            f"name={self.name}, "
            f"path={self.path}, "
            f"description={self.description}, "
            f"commit={self.commit}, "
            f"present_in_specfile={self.present_in_specfile}, "
            f"ignore={self.ignore}, "
            f"squash_commits={self.squash_commits}, "
            f"patch_id={self.patch_id}, "
            f"no_prefix={self.no_prefix}, "
            f"metadata_defined={self.metadata_defined}"
            ")"
        )


class PatchGenerator:
    """
    Generate .patch files from git
    """

    def __init__(self, lp: LocalProject):
        self.lp = lp

    def are_child_commits_contained(self, git_ref: str) -> bool:
        r"""
        Magic begins here.

        `git format-patch` produces patches which cannot be applied when commits,
        in front of a git ref, have parents behind:

        * | | | | | | | |   ea500ac513 (tag: v245) Merge pull request #15...
        |\ \ \ \ \ \ \ \ \
        | * | | | | | | | | 0d5aef3eb5 hwdb: update for v245
        | | |_|_|_|_|_|/ /
        | |/| | | | | | |
        * | | | | | | | | 03985d069b NEWS: final contributor update for v245

        In this example, you can see that ea500 is tagged with 245 and
        is a merge commit. The additional lines mean that child commits of v245
        have parents behind v245 which means that `git format-patch` may
        create patches which may not able able to be applied.

        This method detects the situation described above.

        :param git_ref: git ref to check
        :return: yes if all child commits of the selected git ref are contained
                 within the set of git_ref children commits
        """
        commits = self.get_commits_since_ref(
            git_ref,
            add_upstream_head_commit=True,
            no_merge_commits=False,
        )
        for commit in islice(commits, 1, None):  # 0 = upstream, don't check that one
            for parent_commit in commit.parents:
                if parent_commit not in commits:
                    logger.info(f"Commit {commit!r} has a parent behind {git_ref!r}.")
                    return False
        logger.debug(f"All commits are contained on top of {git_ref!r}.")
        return True

    def linearize_history(self, git_ref: str) -> str:
        r"""
        Transform complex git history into a linear one starting from a selected git ref.

        Returns the name of the linearized branch.

        Change this:
        * | | | | | | | |   ea500ac513 (tag: v245) Merge pull request #15...
        |\ \ \ \ \ \ \ \ \
        | * | | | | | | | | 0d5aef3eb5 hwdb: update for v245
        | | |_|_|_|_|_|/ /
        | |/| | | | | | |
        * | | | | | | | | 03985d069b NEWS: final contributor update for v245

        Into this:
        * 0d5aef3eb5 hwdb: update for v245
        * 03985d069b NEWS: final contributor update for v245
        """
        logger.info(
            "When git history is too complex with merge commits having parents \n"
            "across a wide range, git is known to produce patches which cannot be applied. \n"
            "Therefore we are going to make the history linear on a dedicated branch \n"
            "to make sure the patches will be able to be applied.",
        )
        if self.lp.git_repo.is_dirty():
            raise PackitGitException(
                "The source-git repo is dirty which means we won't be able to do a linear history. "
                "Please commit the changes to resolve the issue. If you are changing the content "
                "of the repository in an action, you can commit those as well.",
            )
        current_time = datetime.datetime.now().strftime(DATETIME_FORMAT)
        initial_branch = self.lp.ref
        target_branch = f"packit-patches-{current_time}"
        logger.info(f"Switch branch to {target_branch!r}.")
        ref = self.lp.create_branch(target_branch)
        ref.checkout()
        target = f"{git_ref}..HEAD"
        logger.debug(f"Linearize history {target}.")
        # https://stackoverflow.com/a/17994534/909579
        # With this command we will rewrite git history of our newly created branch
        # by dropping the merge commits and setting parent commits to those from target branch
        # this means we will drop the reference from which we are merging
        # filter branch passes these to cut:
        #   ` -p 61f3e897f13101f29fb8027e8839498a469ad58e`
        #   ` -p b7cf4b4ef5d0336443f21809b1506bc4a8aa75a9 -p 257188f80ce1a083e3a88b679b898a7...`
        # so we will keep the first parent and drop all the others
        try:
            run_command(
                [
                    "git",
                    "filter-branch",
                    "-f",
                    "--parent-filter",
                    'cut -f 2,3 -d " "',
                    target,
                ],
                # git prints nasty warning when filter-branch is used that it's dangerous
                # this env var prevents it from printing
                env={"FILTER_BRANCH_SQUELCH_WARNING": "1"},
                print_live=True,
                cwd=self.lp.working_dir,
            )
        finally:
            # check out the former branch
            self.lp.checkout_ref(initial_branch)
            # we could also delete the newly created branch,
            # but let's not do that so that user can inspect it
        return target_branch

    def process_patches(
        self,
        patches: dict[str, bytes],
        commits: list[git.Commit],
        destination: str,
        files_to_ignore: Optional[list[str]] = None,
    ) -> list[PatchMetadata]:
        """
        Pair commits (in a source-git repo) with a list patches generated with git-format-patch.

        Pairing is done using commit.hexsha (which is always present in the patch file).

        patch_list (provided List) is then mutated by appending PatchMetadata using
        the paired information: commit and a path to the patch file.

        :param patches: Dict: commit hexsha -> patch content
        :param commits: list of commits we created the patches from
        :param destination: place the patch files here
        :param files_to_ignore: list of files to ignore when creating patches
        """
        patch_list: list[PatchMetadata] = []
        for commit in commits:
            # commit.size doesn't work since even an empty commit is size > 0 (287)
            if not commit.stats.files:
                # this patch is empty - rpmbuild is okay with empty patches (!)
                logger.debug(f"commit {commit} is empty")
                patch = PatchMetadata.from_commit(commit=commit)
                if not patch.present_in_specfile:
                    # it's empty and not present in spec file
                    # nothing to do here
                    continue
                if not patch.name:
                    raise PackitException(
                        f"Empty commit {commit} is referencing a patch which is present in spec"
                        " file but the name is not defined in the commit metadata"
                        " - please define it.",
                    )
                patch_list.append(patch)

                # we need to create it since `git format-patch` won't
                Path(destination).joinpath(patch.name).write_text("")
                logger.info(f"created empty patch {patch.path}")
                continue
            for patch_name, patch_content in patches.items():
                # `git format-patch` usually creates one patch for a merge commit,
                # so some commits won't be covered by a dedicated patch file
                if commit.hexsha.encode() in patch_content:
                    path = Path(patch_name)
                    patch_metadata = PatchMetadata.from_commit(
                        commit=commit,
                        patch_path=path,
                    )

                    if patch_metadata.ignore:
                        logger.debug(
                            f"[IGNORED: {patch_metadata.name}] {commit.summary}",
                        )
                    else:
                        logger.debug(f"[{patch_metadata.name}] {commit.summary}")
                        if patch_metadata.no_prefix:
                            # Sadly, we have work to do, the original patch is no good:
                            # 'format-patch' by default generates patches with prefixes a/ and b/.
                            # 'no-prefix' means we don't want those: we need to delete and re-create
                            # the patch without the prefixes.
                            # https://github.com/packit/dist-git-to-source-git/issues/85#issuecomment-698827925
                            path.unlink()
                            git_f_p_out = git_format_patch(
                                self.lp.working_dir,
                                destination,
                                files_to_ignore,
                                f"{commit}^..{commit}",
                                no_prefix=True,
                            )
                            patch_list.append(
                                PatchMetadata.from_commit(
                                    commit=commit,
                                    patch_path=Path(git_f_p_out),
                                ),
                            )
                        else:
                            patch_list.append(patch_metadata)
                    break
        return patch_list

    def process_patches_with_trailers(
        self,
        patches: list[str],
    ) -> tuple[list[PatchMetadata], bool]:
        """Collect and return patch metadata stored in Git trailers.

        Args:
            patches: List of patch paths as produced by 'git format-patch'
                with the '--output-directory' option.

        Returns:
            A list of PatchMetadata objects and a flag to indicate whether any metadata
            stored in Git trailers was found or not.
        """
        patch_list: list[PatchMetadata] = [
            PatchMetadata.from_patch(patch) for patch in patches
        ]
        used_git_trailers = any(patch.metadata_defined for patch in patch_list)
        if used_git_trailers:
            patch_list = [patch for patch in patch_list if not patch.ignore]
            (remove_prefixes(patch.path) for patch in patch_list if patch.no_prefix)
        return patch_list, used_git_trailers

    @staticmethod
    def squash_by_squash_commits(
        patch_list: list[PatchMetadata],
    ) -> list[PatchMetadata]:
        """Append commits to a single patch until 'squash_commits==True'
        is found.

        Do this after reversing the list of commits, so that they are in a
        newest-to-oldest order, that is HEAD is the first in the list.

        Args:
            patch_list: List of patch metadata objects, corresponding to commits,
                from the oldest to the newest (that is, in the order 'git format-patch'
                produces them)

        Returns:
            List of patch metadata objects, with commits squashed if required.
        """
        logger.warning(
            "Squashing commits by 'squash_commits' metadata. "
            "This mechanism is deprecated. Please, use an "
            "identical 'patch_name' to merge multiple adjacent "
            "commits in a single patch file, instead.",
        )

        new_patch_list: list[PatchMetadata] = []

        prepend_patches = ""
        # this iterator is being reused in both cycles below
        # first we process the commits before first squash_commits and
        # prepend them before the first top_commit
        # while the second cycle goes through the rest of the patches
        reversed_patch_list = reversed(patch_list)
        while True:
            # iterate through the list until squash_commits=True is found
            # prepend those commits to the first top patch
            patch = next(reversed_patch_list)
            if patch.squash_commits:
                # top_patch is a leading commit with squash_commits=True
                top_patch = patch
                break
            logger.debug(f"Prepending patch {patch}.")
            prepend_patches = patch.path.read_text() + prepend_patches
            patch.path.unlink()
        if prepend_patches:
            top_patch.path.write_text(top_patch.path.read_text() + prepend_patches)

        while True:
            if (
                patch.squash_commits
                or patch.present_in_specfile
                or patch.metadata_defined
            ):
                top_patch = patch
                new_patch_list.append(top_patch)
                logger.debug(f"Top commit in a patch: {top_patch}.")
            else:
                logger.debug(f"Appending commit {patch} to {top_patch}.")
                top_patch.path.write_text(
                    patch.path.read_text() + top_patch.path.read_text(),
                )
                patch.path.unlink()
            # we are draining rest of the iterator here
            try:
                patch = next(reversed_patch_list)
            except StopIteration:
                break

        # we need to reverse it back to be consistent with the order
        new_patch_list.reverse()
        return new_patch_list

    @staticmethod
    def squash_by_patch_name(patch_list: list[PatchMetadata]) -> list[PatchMetadata]:
        """Squash adjacent patches if they have identical names.

        Squashing is done by appending the content of a patch to the previous patch
        if the name of the patch matches the name of the previous patch, and removing
        the corresponding PatchMetadata object from the list.

        Note, that renaming the patch files so that their names matches the name
        specified in the metadata is done in 'rename_patches()'.

        Args:
            patch_list: List of PatchMetadata objects, ordered from
                the oldest to the newest commit.

        Returns:
            List of PatchMetadata objects after squashing the patches.

        Raises:
            PackitException if non-adjacent patches have the same name.
        """
        logger.debug("Squashing commits by 'patch_name'.")
        squashed_patch_list: list[PatchMetadata] = []
        seen_patch_names = set()
        for patch in patch_list:
            if squashed_patch_list and squashed_patch_list[-1].name == patch.name:
                logger.debug(
                    f"Appending patch {patch!r} to {squashed_patch_list[-1]!r}.",
                )
                squashed_patch_list[-1].path.write_text(
                    squashed_patch_list[-1].path.read_text() + patch.path.read_text(),
                )
                patch.path.unlink()
            else:
                if patch.name in seen_patch_names:
                    raise PackitException(
                        f"Non-adjacent patches cannot have the same name: {patch.name}.",
                    )
                seen_patch_names.add(patch.name)
                squashed_patch_list.append(patch)
        return squashed_patch_list

    @staticmethod
    def squash_patches(
        patch_list: list[PatchMetadata],
    ) -> list[PatchMetadata]:
        """
        When using `%autosetup -S git_am`, there is a case
        where a single patch file contains multiple commits.
        This is problematic for us since only the leading commit
        is annotated. To make matters worse, we also want to support the fact
        that leading commits would not follow the scheme - make contributions easier.

        In this case, we need to append commits to a single patch, either by looking
        at the 'squash_commits' metadata or by identical 'patch_name'-s, and return
        the new list of patches.

        Args:
            patch_list: List of patch metadata objects, corresponding to commits,
                from the oldest to the newest (that is, in the order 'git format-patch'
                produces them)

        Returns:
            List of patch metadata objects, with commits squashed if required.
        """
        if any(commit.squash_commits for commit in patch_list):
            return PatchGenerator.squash_by_squash_commits(patch_list)
        return PatchGenerator.squash_by_patch_name(patch_list)

    @staticmethod
    def rename_patches(patch_list: list[PatchMetadata]):
        """Ensure that patch file names match 'patch_name', if defined.

        Args:
            patch_list: A list of PatchMetadata objects.
        """
        for patch in patch_list:
            # Not all patches have a name.
            # Empty patches don't have a path.
            if not (patch.name and patch.path):
                continue
            if patch.name != patch.path.name:
                new_path = patch.path.parent / patch.name
                logger.debug(
                    f"Renaming the patch: {patch.path} -> {new_path}"
                    f"{' (already exists)' if new_path.exists() else ''}",
                )
                patch.path.rename(new_path)
                patch.path = new_path

    @staticmethod
    def undo_identical(
        patch_list: list[PatchMetadata],
        repo: git.Repo,
    ) -> list[PatchMetadata]:
        """
        Remove from patch_list and undo changes of patch files which
        have the same patch-id as their previous version, and so they
        can be considerd identical.

        :param patch_list: List of patches to check.
        :param repo: Git repo to work in.
        :return: A filtered list of patches, with identical patches removed.
        """
        ret: list[PatchMetadata] = []
        for patch in patch_list:
            relative_patch_path = patch.path.relative_to(repo.working_dir)
            logger.debug(f"Processing {relative_patch_path} ...")
            new_patch = str(relative_patch_path) in repo.untracked_files
            if new_patch:
                logger.debug(f"{relative_patch_path} is a new patch")
                ret.append(patch)
                continue

            # patch-id before the change
            prev_patch = repo.git.show(f"HEAD:{relative_patch_path}")
            prev_patch = git_patch_ish(prev_patch)
            # We can't know the encoding of the patch.
            # https://docs.python.org/3/howto/unicode.html#files-in-an-unknown-encoding
            with tempfile.TemporaryFile(mode="w+", errors="surrogateescape") as fp:
                fp.write(prev_patch)
                fp.seek(0)
                prev_patch_id = repo.git.patch_id("--stable", istream=fp).split()[0]
                logger.debug(f"Previous patch-id: {prev_patch_id}")

            # current patch-id
            with open(patch.path) as fp:
                current_patch_id = repo.git.patch_id("--stable", istream=fp).split()[0]
                logger.debug(f"Current patch-id: {current_patch_id}")

            if current_patch_id != prev_patch_id:
                ret.append(patch)
            else:
                # this looks the same, don't change it
                repo.git.checkout("--", relative_patch_path)
        return ret

    def create_patches(
        self,
        git_ref: str,
        destination: str,
        files_to_ignore: Optional[list[str]] = None,
    ) -> list[PatchMetadata]:
        """
        Create patches from git commits.

        :param git_ref: start processing commits from this till HEAD
        :param destination: place the patch files here
        :param files_to_ignore: list of files to ignore when creating patches
        :return: [PatchMetadata, ...] list of patches
        """
        files_to_ignore = files_to_ignore or []
        patch_list: list[PatchMetadata] = []
        contained = self.are_child_commits_contained(git_ref)
        patches_revision_range = f"{git_ref}..HEAD"
        if not contained:
            lin_branch = self.linearize_history(git_ref)
            # we're still on the same branch but want to get patches from the linearized branch
            patches_revision_range = f"{git_ref}..{lin_branch}"

        # this is a string, separated by new-lines, with the names of patch files
        git_format_patch_out = git_format_patch(
            self.lp.working_dir,
            destination,
            files_to_ignore,
            patches_revision_range,
        )

        if git_format_patch_out:
            patches: dict[str, bytes] = {
                # we need to read bytes since we cannot decode whatever is inside patches
                patch_name: Path(patch_name).read_bytes()
                for patch_name in git_format_patch_out.split("\n")
            }
            commits = self.get_commits_in_range(patches_revision_range)
            patch_list, used_git_trailers = self.process_patches_with_trailers(
                list(patches),
            )
            if not used_git_trailers:
                patch_list = self.process_patches(
                    patches,
                    commits,
                    destination,
                    files_to_ignore,
                )
            patch_list = self.squash_patches(patch_list)
            self.rename_patches(patch_list)
        else:
            logger.info(f"No patches in range {patches_revision_range}")

        return patch_list

    def get_commits_since_ref(
        self,
        git_ref: str,
        add_upstream_head_commit: bool = True,
        no_merge_commits: bool = True,
    ) -> list[git.Commit]:
        """
        Return a list of different commits between HEAD and selected git_ref

        :param git_ref: get commits since this git ref
        :param add_upstream_head_commit: add also upstream rev/tag commit as a first value
        :param no_merge_commits: do not include merge commits in the list if True
        :return: list of commits (last commit on the current branch)
        """
        if is_a_git_ref(repo=self.lp.git_repo, ref=git_ref):
            upstream_ref = git_ref
        else:
            upstream_ref = f"origin/{git_ref}"
            if upstream_ref not in self.lp.git_repo.refs:
                raise PackitException(
                    f"Couldn't not find upstream branch {upstream_ref!r} and tag {git_ref!r}.",
                )
        commits = self.get_commits_in_range(
            revision_range=f"{git_ref}..{self.lp.ref}",
            no_merge_commits=no_merge_commits,
        )
        if add_upstream_head_commit:
            commits.insert(0, self.lp.git_repo.commit(upstream_ref))

        logger.debug(f"Delta ({upstream_ref}..{self.lp.ref}): {len(commits)}")
        return commits

    def get_commits_in_range(
        self,
        revision_range: str,
        no_merge_commits: bool = True,
    ) -> list[git.Commit]:
        """
        provide a list of git.Commit objects in a given git range

        :param revision_range: e.g. `branch..HEAD`, see `man git-log` for more info
        :param no_merge_commits: don't include merge commits in the list
        :return: list of commits (last commit on the current branch)
        """
        return list(
            self.lp.git_repo.iter_commits(
                rev=revision_range,
                reverse=True,
                no_merges=no_merge_commits,  # do not include merge commits in the list
            ),
        )
